import scipy.stats

import autograd.numpy as np
from autograd.extend import defvjp, primitive
from autograd.scipy.special import digamma

rvs = primitive(scipy.stats.dirichlet.rvs)
pdf = primitive(scipy.stats.dirichlet.pdf)
logpdf = primitive(scipy.stats.dirichlet.logpdf)

defvjp(
    logpdf,
    lambda ans, x, alpha: lambda g: g * (alpha - 1) / x,
    lambda ans, x, alpha: lambda g: g * (digamma(np.sum(alpha)) - digamma(alpha) + np.log(x)),
)

# Same as log pdf, but multiplied by the pdf (ans).
defvjp(
    pdf,
    lambda ans, x, alpha: lambda g: g * ans * (alpha - 1) / x,
    lambda ans, x, alpha: lambda g: g * ans * (digamma(np.sum(alpha)) - digamma(alpha) + np.log(x)),
)



import itertools as it
from functools import reduce

import numpy as np

from autograd.core import vspace
from autograd.test_util import check_grads, scalar_close


def check_vspace(value):
    vs = vspace(value)
    # --- required attributes ---
    size = vs.size
    add = vs.add
    scalar_mul = vs.scalar_mul
    inner_prod = vs.inner_prod
    randn = vs.randn
    zeros = vs.zeros
    ones = vs.ones
    standard_basis = vs.standard_basis

    # --- util ---
    def randns(N=2):
        return [randn() for i in range(N)]

    def rand_scalar():
        return float(np.random.randn())

    def rand_scalars(N=2):
        return [rand_scalar() for i in range(N)]

    def vector_close(x, y):
        z = randn()
        return scalar_close(inner_prod(z, x), inner_prod(z, y))

    # --- vector space axioms ---
    def associativity_of_add(x, y, z):
        return vector_close(add(x, add(y, z)), add(add(x, y), z))

    def commutativity_of_add(x, y):
        return vector_close(add(x, y), add(y, x))

    def identity_element_of_add(x):
        return vector_close(add(zeros(), x), x)

    def inverse_elements_of_add(x):
        return vector_close(zeros(), add(x, scalar_mul(x, -1.0)))

    def compatibility_of_scalar_mul_with_field_mul(x, a, b):
        return vector_close(scalar_mul(x, a * b), scalar_mul(scalar_mul(x, a), b))

    def identity_element_of_scalar_mul(x):
        return vector_close(scalar_mul(x, 1.0), x)

    def distributivity_of_scalar_mul_wrt_vector_add(x, y, a):
        return vector_close(scalar_mul(add(x, y), a), add(scalar_mul(x, a), scalar_mul(y, a)))

    def distributivity_of_scalar_mul_wrt_scalar_add(x, a, b):
        return vector_close(scalar_mul(x, a + b), add(scalar_mul(x, a), scalar_mul(x, b)))

    # --- closure ---
    def add_preserves_vspace(x, y):
        return vs == vspace(add(x, y))

    def scalar_mul_preserves_vspace(x, a):
        return vs == vspace(scalar_mul(x, a))

    # --- inner product axioms ---
    def symmetry(x, y):
        return scalar_close(inner_prod(x, y), inner_prod(y, x))

    def linearity(x, y, a):
        return scalar_close(inner_prod(scalar_mul(x, a), y), a * inner_prod(x, y))

    def positive_definitive(x):
        return 0 < inner_prod(x, x)

    def inner_zeros():
        return scalar_close(0, inner_prod(zeros(), zeros()))

    # --- basis vectors and special vectors---
    def basis_orthonormality():
        return all(
            [
                scalar_close(inner_prod(x, y), 1.0 * (ix == iy))
                for (ix, x), (iy, y) in it.product(enumerate(standard_basis()), enumerate(standard_basis()))
            ]
        )

    def ones_sum_of_basis_vects():
        return vector_close(reduce(add, standard_basis()), ones())

    def basis_correct_size():
        return len(list(standard_basis())) == size

    def basis_correct_vspace():
        return (vs == vspace(x) for x in standard_basis())

    def zeros_correct_vspace():
        return vs == vspace(zeros())

    def ones_correct_vspace():
        return vs == vspace(ones())

    def randn_correct_vspace():
        return vs == vspace(randn())

    assert associativity_of_add(*randns(3))
    assert commutativity_of_add(*randns())
    assert identity_element_of_add(randn())
    assert inverse_elements_of_add(randn())
    assert compatibility_of_scalar_mul_with_field_mul(randn(), *rand_scalars())
    assert identity_element_of_scalar_mul(randn())
    assert distributivity_of_scalar_mul_wrt_vector_add(randn(), randn(), rand_scalar())
    assert distributivity_of_scalar_mul_wrt_scalar_add(randn(), *rand_scalars())
    assert add_preserves_vspace(*randns())
    assert scalar_mul_preserves_vspace(randn(), rand_scalar())
    assert symmetry(*randns())
    assert linearity(randn(), randn(), rand_scalar())
    assert positive_definitive(randn())
    assert inner_zeros()
    assert basis_orthonormality()
    assert ones_sum_of_basis_vects()
    assert basis_correct_size()
    assert basis_correct_vspace()
    assert zeros_correct_vspace()
    assert ones_correct_vspace()
    assert randn_correct_vspace()

    # --- grads of basic operations ---
    check_grads(add)(*randns())
    check_grads(scalar_mul)(randn(), rand_scalar())
    check_grads(inner_prod)(*randns())


def test_array_vspace():
    check_vspace(np.zeros((3, 2)))


def test_array_vspace_0_dim():
    check_vspace(0.0)


def test_array_vspace_complex():
    check_vspace(1.0j * np.zeros((2, 1)))


def test_list_vspace():
    check_vspace([1.0, np.zeros((2, 1))])


def test_tuple_vspace():
    check_vspace((1.0, np.zeros((2, 1))))


def test_dict_vspace():
    check_vspace({"a": 1.0, "b": np.zeros((2, 1))})


def test_mixed_vspace():
    check_vspace({"x": [0.0, np.zeros((3, 1))], "y": ({"a": 0.0}, [0.0])})



import scipy.stats

import autograd.numpy as np
from autograd.extend import defvjp, primitive
from autograd.numpy.numpy_vjps import unbroadcast_f
from autograd.scipy.special import beta, psi

cdf = primitive(scipy.stats.beta.cdf)
logpdf = primitive(scipy.stats.beta.logpdf)
pdf = primitive(scipy.stats.beta.pdf)


def grad_beta_logpdf_arg0(x, a, b):
    return (1 + a * (x - 1) + x * (b - 2)) / (x * (x - 1))


def grad_beta_logpdf_arg1(x, a, b):
    return np.log(x) - psi(a) + psi(a + b)


def grad_beta_logpdf_arg2(x, a, b):
    return np.log1p(-x) - psi(b) + psi(a + b)


defvjp(
    cdf,
    lambda ans, x, a, b: unbroadcast_f(
        x, lambda g: g * np.power(x, a - 1) * np.power(1 - x, b - 1) / beta(a, b)
    ),
    argnums=[0],
)
defvjp(
    logpdf,
    lambda ans, x, a, b: unbroadcast_f(x, lambda g: g * grad_beta_logpdf_arg0(x, a, b)),
    lambda ans, x, a, b: unbroadcast_f(a, lambda g: g * grad_beta_logpdf_arg1(x, a, b)),
    lambda ans, x, a, b: unbroadcast_f(b, lambda g: g * grad_beta_logpdf_arg2(x, a, b)),
)
defvjp(
    pdf,
    lambda ans, x, a, b: unbroadcast_f(x, lambda g: g * ans * grad_beta_logpdf_arg0(x, a, b)),
    lambda ans, x, a, b: unbroadcast_f(a, lambda g: g * ans * grad_beta_logpdf_arg1(x, a, b)),
    lambda ans, x, a, b: unbroadcast_f(b, lambda g: g * ans * grad_beta_logpdf_arg2(x, a, b)),
)



import matplotlib.pyplot as plt
from gaussian_process import make_gp_funs, rbf_covariance
from scipy.optimize import minimize

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import value_and_grad


def build_step_function_dataset(D=1, n_data=40, noise_std=0.1):
    rs = npr.RandomState(0)
    inputs = np.linspace(-2, 2, num=n_data)
    targets = np.sign(inputs) + rs.randn(n_data) * noise_std
    inputs = inputs.reshape((len(inputs), D))
    return inputs, targets


def build_deep_gp(input_dimension, hidden_dimension, covariance_function):
    # GP going from input to hidden
    num_params_layer1, predict_layer1, log_marginal_likelihood_layer1 = make_gp_funs(
        covariance_function, num_cov_params=input_dimension + 1
    )

    # GP going from hidden to output
    num_params_layer2, predict_layer2, log_marginal_likelihood_layer2 = make_gp_funs(
        covariance_function, num_cov_params=hidden_dimension + 1
    )

    num_hidden_params = hidden_dimension * n_data
    total_num_params = num_params_layer1 + num_params_layer2 + num_hidden_params

    def unpack_all_params(all_params):
        layer1_params = all_params[:num_params_layer1]
        layer2_params = all_params[num_params_layer1 : num_params_layer1 + num_params_layer2]
        hiddens = all_params[num_params_layer1 + num_params_layer2 :]
        return layer1_params, layer2_params, hiddens

    def combined_predict_fun(all_params, X, y, xs):
        layer1_params, layer2_params, hiddens = unpack_all_params(all_params)
        h_star_mean, h_star_cov = predict_layer1(layer1_params, X, hiddens, xs)
        y_star_mean, y_star_cov = predict_layer2(
            layer2_params, np.atleast_2d(hiddens).T, y, np.atleast_2d(h_star_mean).T
        )
        return y_star_mean, y_star_cov

    def log_marginal_likelihood(all_params):
        layer1_params, layer2_params, h = unpack_all_params(all_params)
        return log_marginal_likelihood_layer1(layer1_params, X, h) + log_marginal_likelihood_layer2(
            layer2_params, np.atleast_2d(h).T, y
        )

    predict_layer_funcs = [predict_layer1, predict_layer2]

    return (
        total_num_params,
        log_marginal_likelihood,
        combined_predict_fun,
        unpack_all_params,
        predict_layer_funcs,
    )


if __name__ == "__main__":
    n_data = 20
    input_dimension = 1
    hidden_dimension = 1
    X, y = build_step_function_dataset(D=input_dimension, n_data=n_data)

    (
        total_num_params,
        log_marginal_likelihood,
        combined_predict_fun,
        unpack_all_params,
        predict_layer_funcs,
    ) = build_deep_gp(input_dimension, hidden_dimension, rbf_covariance)

    # Set up figure.
    fig = plt.figure(figsize=(12, 8), facecolor="white")
    ax_end_to_end = fig.add_subplot(311, frameon=False)
    ax_x_to_h = fig.add_subplot(312, frameon=False)
    ax_h_to_y = fig.add_subplot(313, frameon=False)
    plt.show(block=False)

    def plot_gp(ax, X, y, pred_mean, pred_cov, plot_xs):
        ax.cla()
        marg_std = np.sqrt(np.diag(pred_cov))
        ax.plot(plot_xs, pred_mean, "b")
        ax.fill(
            np.concatenate([plot_xs, plot_xs[::-1]]),
            np.concatenate([pred_mean - 1.96 * marg_std, (pred_mean + 1.96 * marg_std)[::-1]]),
            alpha=0.15,
            fc="Blue",
            ec="None",
        )

        # Show samples from posterior.
        rs = npr.RandomState(0)
        sampled_funcs = rs.multivariate_normal(pred_mean, pred_cov, size=10)
        ax.plot(plot_xs, sampled_funcs.T)
        ax.plot(X, y, "kx")
        ax.set_ylim([-1.5, 1.5])
        ax.set_xticks([])
        ax.set_yticks([])

    def callback(params):
        print(f"Log marginal likelihood {log_marginal_likelihood(params)}")

        # Show posterior marginals.
        plot_xs = np.reshape(np.linspace(-5, 5, 300), (300, 1))
        pred_mean, pred_cov = combined_predict_fun(params, X, y, plot_xs)
        plot_gp(ax_end_to_end, X, y, pred_mean, pred_cov, plot_xs)
        ax_end_to_end.set_title("X to y")

        layer1_params, layer2_params, hiddens = unpack_all_params(params)
        h_star_mean, h_star_cov = predict_layer_funcs[0](layer1_params, X, hiddens, plot_xs)
        y_star_mean, y_star_cov = predict_layer_funcs[0](layer2_params, np.atleast_2d(hiddens).T, y, plot_xs)

        plot_gp(ax_x_to_h, X, hiddens, h_star_mean, h_star_cov, plot_xs)
        ax_x_to_h.set_title("X to hiddens")

        plot_gp(ax_h_to_y, np.atleast_2d(hiddens).T, y, y_star_mean, y_star_cov, plot_xs)
        ax_h_to_y.set_title("hiddens to y")

        plt.draw()
        plt.pause(1.0 / 60.0)

    # Initialize covariance parameters and hiddens.
    rs = npr.RandomState(0)
    init_params = 0.1 * rs.randn(total_num_params)

    print("Optimizing covariance parameters...")
    objective = lambda params: -log_marginal_likelihood(params)
    cov_params = minimize(value_and_grad(objective), init_params, jac=True, method="CG", callback=callback)
    plt.pause(10.0)



import time
import warnings

import autograd.numpy as np
from autograd import deriv, grad
from autograd.builtins import list as make_list


def timefunction(f):
    t = time.time()
    f()
    return time.time() - t


def assert_linear_time(f):
    t = timefunction(lambda: f(1))
    t10 = timefunction(lambda: f(10))
    assert t10 > 5 * t, f"Too fast: f(1) takes {t}, f(10) takes {t10}"
    assert t10 < 20 * t, f"Too slow: f(1) takes {t}, f(10) takes {t10}"
    if not (8 * t < t10 < 12 * t):
        warnings.warn("Borderline linearity. May fail on different hardware")


def test_array_creation():
    def fun(x, N):
        arr = [x for i in range(N)]
        return np.sum(np.array(arr))

    assert_linear_time(lambda N: grad(fun)(1.0, 200 * N))


def test_array_indexing():
    def fun(x):
        return sum([x[i] for i in range(len(x))])

    assert_linear_time(lambda N: grad(fun)(np.zeros(200 * N)))


def test_list_indexing():
    def fun(x):
        return sum([x[i] for i in range(len(x))])

    assert_linear_time(lambda N: grad(fun)([0.0 for i in range(50 * N)]))


def test_list_creation():
    def fun(x, N):
        return make_list(*[x for _ in range(N)])

    assert_linear_time(lambda N: deriv(fun)(0.0, 20 * N))


# This fails. Need to figure out why
def test_array_creation_fwd():
    def fun(x, N):
        arr = [x for i in range(N)]
        return np.sum(np.array(arr))

    assert_linear_time(lambda N: deriv(fun)(1.0, 400 * N))



import os

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.pyplot import imread
from scipy.optimize import minimize

import autograd.numpy as np
from autograd import value_and_grad

# Fluid simulation code based on
# "Real-Time Fluid Dynamics for Games" by Jos Stam
# https://www.josstam.com/_files/ugd/cf1fd6_9989229efbd34a26ba5ccd913721a2ac.pdf


def project(vx, vy):
    """Project the velocity field to be approximately mass-conserving,
    using a few iterations of Gauss-Seidel."""
    p = np.zeros(vx.shape)
    h = 1.0 / vx.shape[0]
    div = (
        -0.5
        * h
        * (
            np.roll(vx, -1, axis=0)
            - np.roll(vx, 1, axis=0)
            + np.roll(vy, -1, axis=1)
            - np.roll(vy, 1, axis=1)
        )
    )

    for k in range(10):
        p = (
            div
            + np.roll(p, 1, axis=0)
            + np.roll(p, -1, axis=0)
            + np.roll(p, 1, axis=1)
            + np.roll(p, -1, axis=1)
        ) / 4.0

    vx -= 0.5 * (np.roll(p, -1, axis=0) - np.roll(p, 1, axis=0)) / h
    vy -= 0.5 * (np.roll(p, -1, axis=1) - np.roll(p, 1, axis=1)) / h
    return vx, vy


def advect(f, vx, vy):
    """Move field f according to x and y velocities (u and v)
    using an implicit Euler integrator."""
    rows, cols = f.shape
    cell_ys, cell_xs = np.meshgrid(np.arange(rows), np.arange(cols))
    center_xs = (cell_xs - vx).ravel()
    center_ys = (cell_ys - vy).ravel()

    # Compute indices of source cells.
    left_ix = np.floor(center_xs).astype(int)
    top_ix = np.floor(center_ys).astype(int)
    rw = center_xs - left_ix  # Relative weight of right-hand cells.
    bw = center_ys - top_ix  # Relative weight of bottom cells.
    left_ix = np.mod(left_ix, rows)  # Wrap around edges of simulation.
    right_ix = np.mod(left_ix + 1, rows)
    top_ix = np.mod(top_ix, cols)
    bot_ix = np.mod(top_ix + 1, cols)

    # A linearly-weighted sum of the 4 surrounding cells.
    flat_f = (1 - rw) * ((1 - bw) * f[left_ix, top_ix] + bw * f[left_ix, bot_ix]) + rw * (
        (1 - bw) * f[right_ix, top_ix] + bw * f[right_ix, bot_ix]
    )
    return np.reshape(flat_f, (rows, cols))


def simulate(vx, vy, smoke, num_time_steps, ax=None, render=False):
    print("Running simulation...")
    for t in range(num_time_steps):
        if ax:
            plot_matrix(ax, smoke, t, render)
        vx_updated = advect(vx, vx, vy)
        vy_updated = advect(vy, vx, vy)
        vx, vy = project(vx_updated, vy_updated)
        smoke = advect(smoke, vx, vy)
    if ax:
        plot_matrix(ax, smoke, num_time_steps, render)
    return smoke


def plot_matrix(ax, mat, t, render=False):
    plt.cla()
    ax.matshow(mat)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.draw()
    if render:
        matplotlib.image.imsave(f"step{t:03d}.png", mat)
    plt.pause(0.001)


if __name__ == "__main__":
    simulation_timesteps = 100
    basepath = os.path.dirname(__file__)

    print("Loading initial and target states...")
    init_smoke = imread(os.path.join(basepath, "init_smoke.png"))[:, :, 0]
    # target = imread('peace.png')[::2,::2,3]
    target = imread(os.path.join(basepath, "skull.png"))[::2, ::2]
    rows, cols = target.shape

    init_dx_and_dy = np.zeros((2, rows, cols)).ravel()

    def distance_from_target_image(smoke):
        return np.mean((target - smoke) ** 2)

    def convert_param_vector_to_matrices(params):
        vx = np.reshape(params[: (rows * cols)], (rows, cols))
        vy = np.reshape(params[(rows * cols) :], (rows, cols))
        return vx, vy

    def objective(params):
        init_vx, init_vy = convert_param_vector_to_matrices(params)
        final_smoke = simulate(init_vx, init_vy, init_smoke, simulation_timesteps)
        return distance_from_target_image(final_smoke)

    # Specify gradient of objective function using autograd.
    objective_with_grad = value_and_grad(objective)

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, frameon=False)

    def callback(params):
        init_vx, init_vy = convert_param_vector_to_matrices(params)
        simulate(init_vx, init_vy, init_smoke, simulation_timesteps, ax)

    print("Optimizing initial conditions...")
    result = minimize(
        objective_with_grad,
        init_dx_and_dy,
        jac=True,
        method="CG",
        options={"maxiter": 25, "disp": True},
        callback=callback,
    )

    print("Rendering optimized flow...")
    init_vx, init_vy = convert_param_vector_to_matrices(result.x)
    simulate(init_vx, init_vy, init_smoke, simulation_timesteps, ax, render=True)

    print("Converting frames to an animated GIF...")
    os.system("convert -delay 5 -loop 0 step*.png -delay 250 step100.png surprise.gif")  # Using imagemagick.
    os.system("rm step*.png")



import numpy as onp

import autograd.numpy as np
from autograd import grad

try:
    from autograd.core import VJPNode, backward_pass, vspace
    from autograd.tracer import new_box, trace

    MASTER_BRANCH = False
except ImportError:
    from autograd.core import backward_pass, forward_pass, new_progenitor, vspace

    MASTER_BRANCH = True


## SHORT FUNCTION
def f_short(x):
    return x**2


def time_short_fun():
    f_short(2.0)


def time_short_forward_pass():
    if MASTER_BRANCH:
        forward_pass(f_short, (2.0,), {})
    else:
        start_node = VJPNode.new_root()
        trace(start_node, f_short, x)


def time_short_backward_pass():
    if MASTER_BRANCH:
        backward_pass(1.0, short_end_node, short_start_node)
    else:
        backward_pass(1.0, short_end_node)


def time_short_grad():
    grad(f_short)(2.0)


## LONG FUNCTION
def f_long(x):
    for i in range(50):
        x = np.sin(x)
    return x


def time_long_fun():
    f_long(2.0)


def time_long_forward_pass():
    if MASTER_BRANCH:
        forward_pass(f_long, (2.0,), {})
    else:
        start_node = VJPNode.new_root()
        trace(start_node, f_long, x)


def time_long_backward_pass():
    if MASTER_BRANCH:
        backward_pass(1.0, long_end_node, long_start_node)
    else:
        backward_pass(1.0, long_end_node)


def time_long_grad():
    grad(f_long)(2.0)


## 'PEARLMUTTER TEST' FUNCTION
def fan_out_fan_in(x):
    for i in range(10**4):
        x = (x + x) / 2.0
    return np.sum(x)


def time_fan_out_fan_in_fun():
    fan_out_fan_in(2.0)


def time_fan_out_fan_in_forward_pass():
    if MASTER_BRANCH:
        forward_pass(fan_out_fan_in, (2.0,), {})
    else:
        start_node = VJPNode.new_root()
        trace(start_node, fan_out_fan_in, x)


def time_fan_out_fan_in_backward_pass():
    if MASTER_BRANCH:
        backward_pass(1.0, fan_end_node, fan_start_node)
    else:
        backward_pass(1.0, fan_end_node)


def time_fan_out_fan_in_grad():
    grad(fan_out_fan_in)(2.0)


## UNIT BENCHMARKS
def time_vspace_float():
    vspace(1.0)


A = np.array([[1.0, 2.0, 3.0]])


def time_vspace_array():
    vspace(A)


def time_new_box_float():
    new_box(1.0, 0, start_node)


def time_new_box_array():
    new_box(A, 0, start_node)


def time_exp_call():
    onp.exp(2.0)


def time_exp_primitive_call_unboxed():
    np.exp(2.0)


def time_exp_primitive_call_boxed():
    if MASTER_BRANCH:
        np.exp(progenitor)
    else:
        np.exp(start_box)


def time_no_autograd_control():
    # Test whether the benchmarking machine is running slowly independent of autograd
    A = np.random.randn(200, 200)
    np.dot(A, A)


if MASTER_BRANCH:
    short_start_node, short_end_node = forward_pass(f_short, (2.0,), {})
    long_start_node, long_end_node = forward_pass(f_long, (2.0,), {})
    fan_start_node, fan_end_node = forward_pass(fan_out_fan_in, (2.0,), {})
    progenitor = new_progenitor(2.0)
else:
    x = 2.0
    start_node = VJPNode.new_root()
    start_box = new_box(x, 0, start_node)
    _, short_end_node = trace(VJPNode.new_root(), f_short, x)
    _, long_end_node = trace(VJPNode.new_root(), f_long, x)
    _, fan_end_node = trace(VJPNode.new_root(), fan_out_fan_in, x)



import warnings

import pytest

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad
from autograd.test_util import check_grads

npr.seed(1)


def test_grad_fanout():
    fun = lambda x: np.sin(np.sin(x) + np.sin(x))
    df = grad(fun)
    check_grads(fun)(npr.randn())
    check_grads(df)(npr.rand())


def test_grad_const():
    fun = lambda x: 1.0
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("ignore")
        df = grad(fun)
        assert np.allclose(df(2.0), 0.0)


def test_grad_identity():
    fun = lambda x: x
    df = grad(fun)
    ddf = grad(df)
    assert np.allclose(df(2.0), 1.0)
    assert np.allclose(ddf(2.0), 0.0)


def test_hess_vector_prod():
    npr.seed(1)
    randv = npr.randn(10)

    def fun(x):
        return np.sin(np.dot(x, randv))

    df = grad(fun)

    def vector_product(x, v):
        return np.sin(np.dot(v, df(x)))

    ddf = grad(vector_product)
    A = npr.randn(10)
    B = npr.randn(10)
    check_grads(fun)(A)
    check_grads(vector_product)(A, B)


def test_enclosing_scope_ref():
    def fun(x):
        inner_fun = lambda y: x * y
        return x * grad(inner_fun)(2.0)

    check_grads(fun)(1.0)


def test_enclosing_scope_ref_2():
    def fun(x):
        inner_fun = lambda y: y * x
        return x * grad(inner_fun)(2.0)

    check_grads(fun)(1.0)


def test_mutating_outgrad():
    def fun(a):
        b = a + 1.0
        c = b + 1.5
        d = a + b
        e = d + c
        return e

    A = npr.randn(5)
    check_grads(fun)(A)


def test_mutating_outgrad_from_indexing():
    def fun(a):
        b = a + 1.0
        c = b[0] + 1.5
        d = a + b
        e = d + c
        return e

    A = npr.randn(5)
    check_grads(fun)(A)


def test_complex_mutating_outgrad_from_indexing():
    def fun(a):
        b = a + 1.0j
        c = b[0] + 1.5
        d = a + b
        e = d + c
        return np.sum(np.sin(np.real(e)))

    A = npr.randn(5)
    check_grads(fun)(A)
    d_fun = lambda x: grad(fun)(x)
    check_grads(d_fun)(A)


def test_complex_separate_real_and_imaginary():
    def fun(a):
        r, i = np.real(a), np.imag(a)
        a = np.abs(r) ** 1.4 + np.abs(i) ** 1.3
        return np.sum(np.sin(a))

    d_fun = lambda x: grad(fun)(x)
    A = npr.randn(5, 3) + 0.1j * npr.randn(5, 3)
    check_grads(fun)(A)
    check_grads(d_fun)(A)


def test_third_derivative():
    fun = lambda x: np.sin(np.sin(x) + np.sin(x))
    df = grad(fun)
    ddf = grad(fun)
    dddf = grad(fun)
    check_grads(fun)(npr.randn())
    check_grads(df)(npr.rand())
    check_grads(ddf)(npr.rand())
    check_grads(dddf)(npr.rand())


def test_third_derivative_other_args():
    fun = lambda x, y: np.sin(np.sin(x) + np.sin(y))
    df = grad(fun)
    ddf = grad(fun, 1)
    dddf = grad(fun)
    check_grads(fun)(npr.randn(), npr.randn())
    check_grads(df)(npr.randn(), npr.randn())
    check_grads(ddf)(npr.randn(), npr.randn())
    check_grads(dddf)(npr.randn(), npr.randn())


def test_third_derivative_other_args2():
    fun = lambda x, y: np.sin(np.sin(x) + np.sin(y))
    df = grad(fun, 1)
    ddf = grad(fun)
    dddf = grad(fun, 1)
    check_grads(fun)(npr.randn(), npr.randn())
    check_grads(df)(npr.randn(), npr.randn())
    check_grads(ddf)(npr.randn(), npr.randn())
    check_grads(dddf)(npr.randn(), npr.randn())


def test_singleton_array_output():
    fun = lambda x: np.sum(np.sin(x), keepdims=True)
    check_grads(fun)(npr.randn(3, 3))
    check_grads(lambda x: np.sum(grad(fun)(x)))(npr.randn(3, 3))


def test_singleton_array_output_axis0():
    fun = lambda x: np.sum(np.sin(x), axis=0, keepdims=False)
    check_grads(fun)(npr.randn(3, 1))
    check_grads(lambda x: np.sum(grad(fun)(x)))(npr.randn(3, 1))


def test_singleton_array_output_axis1():
    fun = lambda x: np.sum(np.sin(x), axis=1, keepdims=False)
    check_grads(fun)(npr.randn(1, 3))
    check_grads(lambda x: np.sum(grad(fun)(x)))(npr.randn(1, 3))


def test_singleton_array_output_axis0_keepdims():
    fun = lambda x: np.sum(np.sin(x), axis=0, keepdims=True)
    check_grads(fun)(npr.randn(3, 1))
    check_grads(lambda x: np.sum(grad(fun)(x)))(npr.randn(3, 1))


def test_singleton_array_output_axis1_keepdims():
    fun = lambda x: np.sum(np.sin(x), axis=1, keepdims=True)
    check_grads(fun)(npr.randn(1, 3))
    check_grads(lambda x: np.sum(grad(fun)(x)))(npr.randn(1, 3))


def test_assignment_raises_error():
    def fun(A, b):
        A[1] = b
        return A

    A = npr.randn(5)
    with pytest.raises(TypeError):
        check_grads(fun)(A, 3.0)


# def test_nonscalar_output_1():
#     with pytest.raises(TypeError):
#         grad(lambda x: x * 2)(np.zeros(2))

# def test_nonscalar_output_2():
#     with pytest.raises(TypeError):
#         grad(lambda x: x * 2)(np.zeros(2))

# TODO:
# Diamond patterns
# Taking grad again after returning const
# Empty functions
# 2nd derivatives with fanout, thinking about the outgrad adder



import warnings
from contextlib import contextmanager

import pytest

import autograd.numpy as np
from autograd import deriv, grad
from autograd.core import primitive_vjps
from autograd.extend import primitive
from autograd.test_util import check_grads


def test_assert():
    # from https://github.com/HIPS/autograd/issues/43
    def fun(x):
        assert np.allclose(x, (x * 3.0) / 3.0)
        return np.sum(x)

    check_grads(fun)(np.array([1.0, 2.0, 3.0]))


def test_nograd():
    # we want this to raise non-differentiability error
    fun = lambda x: np.allclose(x, (x * 3.0) / 3.0)
    with pytest.raises(TypeError):
        with warnings.catch_warnings(record=True) as w:
            grad(fun)(np.array([1.0, 2.0, 3.0]))


def test_no_vjp_def():
    fun = primitive(lambda x: 2.0 * x)
    with pytest.raises(NotImplementedError):
        grad(fun)(1.0)


def test_no_jvp_def():
    fun = primitive(lambda x: 2.0 * x)
    with pytest.raises(NotImplementedError):
        deriv(fun)(1.0)


def test_falseyness():
    fun = lambda x: np.real(x**2 if np.iscomplex(x) else np.sum(x))
    check_grads(fun)(2.0)
    check_grads(fun)(2.0 + 1j)


def test_unimplemented_falseyness():
    @contextmanager
    def remove_grad_definitions(fun):
        vjpmaker = primitive_vjps.pop(fun, None)
        yield
        if vjpmaker:
            primitive_vjps[fun] = vjpmaker

    with remove_grad_definitions(np.iscomplex):
        fun = lambda x: np.real(x**2 if np.iscomplex(x) else np.sum(x))
        check_grads(fun)(5.0)
        check_grads(fun)(2.0 + 1j)



import os

import matplotlib.pyplot as plt
from scipy.optimize import minimize

import autograd.numpy as np
from autograd import value_and_grad

rows, cols = 40, 60

# Fluid simulation code based on
# "Real-Time Fluid Dynamics for Games" by Jos Stam
# http://www.intpowertechcorp.com/GDC03.pdf


def occlude(f, occlusion):
    return f * (1 - occlusion)


def project(vx, vy, occlusion):
    """Project the velocity field to be approximately mass-conserving,
    using a few iterations of Gauss-Seidel."""
    p = np.zeros(vx.shape)
    div = -0.5 * (
        np.roll(vx, -1, axis=1) - np.roll(vx, 1, axis=1) + np.roll(vy, -1, axis=0) - np.roll(vy, 1, axis=0)
    )
    div = make_continuous(div, occlusion)

    for k in range(50):
        p = (
            div
            + np.roll(p, 1, axis=1)
            + np.roll(p, -1, axis=1)
            + np.roll(p, 1, axis=0)
            + np.roll(p, -1, axis=0)
        ) / 4.0
        p = make_continuous(p, occlusion)

    vx = vx - 0.5 * (np.roll(p, -1, axis=1) - np.roll(p, 1, axis=1))
    vy = vy - 0.5 * (np.roll(p, -1, axis=0) - np.roll(p, 1, axis=0))

    vx = occlude(vx, occlusion)
    vy = occlude(vy, occlusion)
    return vx, vy


def advect(f, vx, vy):
    """Move field f according to x and y velocities (u and v)
    using an implicit Euler integrator."""
    rows, cols = f.shape
    cell_xs, cell_ys = np.meshgrid(np.arange(cols), np.arange(rows))
    center_xs = (cell_xs - vx).ravel()
    center_ys = (cell_ys - vy).ravel()

    # Compute indices of source cells.
    left_ix = np.floor(center_ys).astype(int)
    top_ix = np.floor(center_xs).astype(int)
    rw = center_ys - left_ix  # Relative weight of right-hand cells.
    bw = center_xs - top_ix  # Relative weight of bottom cells.
    left_ix = np.mod(left_ix, rows)  # Wrap around edges of simulation.
    right_ix = np.mod(left_ix + 1, rows)
    top_ix = np.mod(top_ix, cols)
    bot_ix = np.mod(top_ix + 1, cols)

    # A linearly-weighted sum of the 4 surrounding cells.
    flat_f = (1 - rw) * ((1 - bw) * f[left_ix, top_ix] + bw * f[left_ix, bot_ix]) + rw * (
        (1 - bw) * f[right_ix, top_ix] + bw * f[right_ix, bot_ix]
    )
    return np.reshape(flat_f, (rows, cols))


def make_continuous(f, occlusion):
    non_occluded = 1 - occlusion
    num = (
        np.roll(f, 1, axis=0) * np.roll(non_occluded, 1, axis=0)
        + np.roll(f, -1, axis=0) * np.roll(non_occluded, -1, axis=0)
        + np.roll(f, 1, axis=1) * np.roll(non_occluded, 1, axis=1)
        + np.roll(f, -1, axis=1) * np.roll(non_occluded, -1, axis=1)
    )
    den = (
        np.roll(non_occluded, 1, axis=0)
        + np.roll(non_occluded, -1, axis=0)
        + np.roll(non_occluded, 1, axis=1)
        + np.roll(non_occluded, -1, axis=1)
    )
    return f * non_occluded + (1 - non_occluded) * num / (den + 0.001)


def sigmoid(x):
    return 0.5 * (np.tanh(x) + 1.0)  # Output ranges from 0 to 1.


def simulate(vx, vy, num_time_steps, occlusion, ax=None, render=False):
    occlusion = sigmoid(occlusion)

    # Disallow occlusion outside a certain area.
    mask = np.zeros((rows, cols))
    mask[10:30, 10:30] = 1.0
    occlusion = occlusion * mask

    # Initialize smoke bands.
    red_smoke = np.zeros((rows, cols))
    red_smoke[rows // 4 : rows // 2] = 1
    blue_smoke = np.zeros((rows, cols))
    blue_smoke[rows // 2 : 3 * rows // 4] = 1

    print("Running simulation...")
    vx, vy = project(vx, vy, occlusion)
    for t in range(num_time_steps):
        plot_matrix(ax, red_smoke, occlusion, blue_smoke, t, render)
        vx_updated = advect(vx, vx, vy)
        vy_updated = advect(vy, vx, vy)
        vx, vy = project(vx_updated, vy_updated, occlusion)
        red_smoke = advect(red_smoke, vx, vy)
        red_smoke = occlude(red_smoke, occlusion)
        blue_smoke = advect(blue_smoke, vx, vy)
        blue_smoke = occlude(blue_smoke, occlusion)
    plot_matrix(ax, red_smoke, occlusion, blue_smoke, num_time_steps, render)
    return vx, vy


def plot_matrix(ax, r, g, b, t, render=False):
    if ax:
        plt.cla()
        ax.imshow(np.concatenate((r[..., np.newaxis], g[..., np.newaxis], b[..., np.newaxis]), axis=2))
        ax.set_xticks([])
        ax.set_yticks([])
        plt.draw()
        if render:
            plt.savefig(f"step{t:03d}.png", bbox_inches="tight")
        plt.pause(0.001)


if __name__ == "__main__":
    simulation_timesteps = 20

    print("Loading initial and target states...")
    init_vx = np.ones((rows, cols))
    init_vy = np.zeros((rows, cols))

    # Initialize the occlusion to be a block.
    init_occlusion = -np.ones((rows, cols))
    init_occlusion[15:25, 15:25] = 0.0
    init_occlusion = init_occlusion.ravel()

    def drag(vx):
        return np.mean(init_vx - vx)

    def lift(vy):
        return np.mean(vy - init_vy)

    def objective(params):
        cur_occlusion = np.reshape(params, (rows, cols))
        final_vx, final_vy = simulate(init_vx, init_vy, simulation_timesteps, cur_occlusion)
        return -lift(final_vy) / drag(final_vx)

    # Specify gradient of objective function using autograd.
    objective_with_grad = value_and_grad(objective)

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, frameon=False)

    def callback(weights):
        cur_occlusion = np.reshape(weights, (rows, cols))
        simulate(init_vx, init_vy, simulation_timesteps, cur_occlusion, ax)

    print("Rendering initial flow...")
    callback(init_occlusion)

    print("Optimizing initial conditions...")
    result = minimize(
        objective_with_grad,
        init_occlusion,
        jac=True,
        method="CG",
        options={"maxiter": 50, "disp": True},
        callback=callback,
    )

    print("Rendering optimized flow...")
    final_occlusion = np.reshape(result.x, (rows, cols))
    simulate(init_vx, init_vy, simulation_timesteps, final_occlusion, ax, render=True)

    print("Converting frames to an animated GIF...")  # Using imagemagick.
    os.system(
        "convert -delay 5 -loop 0 step*.png -delay 250 step{:03d}.png wing.gif".format(simulation_timesteps)
    )
    os.system("rm step*.png")



import autograd.numpy as np
from autograd import grad
from autograd.misc.fixed_points import fixed_point


def newton_sqrt_iter(a):
    return lambda x: 0.5 * (x + a / x)


def grad_descent_sqrt_iter(a):
    return lambda x: x - 0.05 * (x**2 - a)


def sqrt(a, guess=10.0):
    # return fixed_point(newton_sqrt_iter, a, guess, distance, 1e-4)
    return fixed_point(grad_descent_sqrt_iter, a, guess, distance, 1e-4)


def distance(x, y):
    return np.abs(x - y)


print(np.sqrt(2.0))
print(sqrt(2.0))
print()
print(grad(np.sqrt)(2.0))
print(grad(sqrt)(2.0))
print()
print(grad(grad(np.sqrt))(2.0))
print(grad(grad(sqrt))(2.0))
print()



import numpy as np

from autograd.builtins import NamedTupleVSpace
from autograd.extend import VSpace


class ArrayVSpace(VSpace):
    def __init__(self, value):
        value = np.asarray(value)
        self.shape = value.shape
        self.dtype = value.dtype

    @property
    def size(self):
        return np.prod(self.shape)

    @property
    def ndim(self):
        return len(self.shape)

    def zeros(self):
        return np.zeros(self.shape, dtype=self.dtype)

    def ones(self):
        return np.ones(self.shape, dtype=self.dtype)

    def standard_basis(self):
        for idxs in np.ndindex(*self.shape):
            vect = np.zeros(self.shape, dtype=self.dtype)
            vect[idxs] = 1
            yield vect

    def randn(self):
        return np.array(np.random.randn(*self.shape)).astype(self.dtype)

    def _inner_prod(self, x, y):
        return np.dot(np.ravel(x), np.ravel(y))


class ComplexArrayVSpace(ArrayVSpace):
    iscomplex = True

    @property
    def size(self):
        return np.prod(self.shape) * 2

    def ones(self):
        return np.ones(self.shape, dtype=self.dtype) + 1.0j * np.ones(self.shape, dtype=self.dtype)

    def standard_basis(self):
        for idxs in np.ndindex(*self.shape):
            for v in [1.0, 1.0j]:
                vect = np.zeros(self.shape, dtype=self.dtype)
                vect[idxs] = v
                yield vect

    def randn(self):
        return np.array(np.random.randn(*self.shape)).astype(self.dtype) + 1.0j * np.array(
            np.random.randn(*self.shape)
        ).astype(self.dtype)

    def _inner_prod(self, x, y):
        return np.real(np.dot(np.conj(np.ravel(x)), np.ravel(y)))

    def _covector(self, x):
        return np.conj(x)


VSpace.register(np.ndarray, lambda x: ComplexArrayVSpace(x) if np.iscomplexobj(x) else ArrayVSpace(x))

for type_ in [float, np.longdouble, np.float64, np.float32, np.float16]:
    ArrayVSpace.register(type_)

for type_ in [complex, np.clongdouble, np.complex64, np.complex128]:
    ComplexArrayVSpace.register(type_)


if np.lib.NumpyVersion(np.__version__) >= "2.0.0":

    class EigResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg._linalg.EigResult

    class EighResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg._linalg.EighResult

    class QRResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg._linalg.QRResult

    class SlogdetResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg._linalg.SlogdetResult

    class SVDResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg._linalg.SVDResult

    EigResultVSpace.register(np.linalg._linalg.EigResult)
    EighResultVSpace.register(np.linalg._linalg.EighResult)
    QRResultVSpace.register(np.linalg._linalg.QRResult)
    SlogdetResultVSpace.register(np.linalg._linalg.SlogdetResult)
    SVDResultVSpace.register(np.linalg._linalg.SVDResult)
elif np.__version__ >= "1.25":

    class EigResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg.linalg.EigResult

    class EighResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg.linalg.EighResult

    class QRResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg.linalg.QRResult

    class SlogdetResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg.linalg.SlogdetResult

    class SVDResultVSpace(NamedTupleVSpace):
        seq_type = np.linalg.linalg.SVDResult

    EigResultVSpace.register(np.linalg.linalg.EigResult)
    EighResultVSpace.register(np.linalg.linalg.EighResult)
    QRResultVSpace.register(np.linalg.linalg.QRResult)
    SlogdetResultVSpace.register(np.linalg.linalg.SlogdetResult)
    SVDResultVSpace.register(np.linalg.linalg.SVDResult)



"""
Inferring a function from a reproducing kernel Hilbert space (RKHS) by taking
gradients of eval with respect to the function-valued argument
"""

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad
from autograd.extend import Box, VSpace, defvjp, primitive
from autograd.util import func


class RKHSFun:
    def __init__(self, kernel, alphas={}):
        self.alphas = alphas
        self.kernel = kernel
        self.vs = RKHSFunVSpace(self)

    @primitive
    def __call__(self, x):
        return sum([a * self.kernel(x, x_repr) for x_repr, a in self.alphas.items()], 0.0)

    def __add__(self, f):
        return self.vs.add(self, f)

    def __mul__(self, a):
        return self.vs.scalar_mul(self, a)


# TODO: add vjp of __call__ wrt x (and show it in action)
defvjp(func(RKHSFun.__call__), lambda ans, f, x: lambda g: RKHSFun(f.kernel, {x: 1}) * g)


class RKHSFunBox(Box, RKHSFun):
    @property
    def kernel(self):
        return self._value.kernel


RKHSFunBox.register(RKHSFun)


class RKHSFunVSpace(VSpace):
    def __init__(self, value):
        self.kernel = value.kernel

    def zeros(self):
        return RKHSFun(self.kernel)

    def randn(self):
        # These arbitrary vectors are not analogous to randn in any meaningful way
        N = npr.randint(1, 3)
        return RKHSFun(self.kernel, dict(zip(npr.randn(N), npr.randn(N))))

    def _add(self, f, g):
        assert f.kernel is g.kernel
        return RKHSFun(f.kernel, add_dicts(f.alphas, g.alphas))

    def _scalar_mul(self, f, a):
        return RKHSFun(f.kernel, {x: a * a_cur for x, a_cur in f.alphas.items()})

    def _inner_prod(self, f, g):
        assert f.kernel is g.kernel
        return sum(
            [a1 * a2 * f.kernel(x1, x2) for x1, a1 in f.alphas.items() for x2, a2 in g.alphas.items()], 0.0
        )


RKHSFunVSpace.register(RKHSFun)


def add_dicts(d1, d2):
    d = {}
    for k, v in d1.items() + d2.items():
        d[k] = d[k] + v if k in d else v
    return d


if __name__ == "__main__":

    def sq_exp_kernel(x1, x2):
        return np.exp(-((x1 - x2) ** 2))

    xs = range(5)
    ys = [1, 2, 3, 2, 1]

    def logprob(f, xs, ys):
        return -sum((f(x) - y) ** 2 for x, y in zip(xs, ys))

    f = RKHSFun(sq_exp_kernel)
    for i in range(100):
        f = f + grad(logprob)(f, xs, ys) * 0.01

    for x, y in zip(xs, ys):
        print(f"{x}\t{y}\t{f(x)}")



# Exposes API for extending autograd
from .core import (
    JVPNode,
    SparseObject,
    VJPNode,
    VSpace,
    def_linear,
    defjvp,
    defjvp_argnum,
    defjvp_argnums,
    defvjp,
    defvjp_argnum,
    defvjp_argnums,
    vspace,
)
from .tracer import Box, notrace_primitive, primitive, register_notrace



# Implements a Gaussian process latent-variable model.
# The (high-dimensional) data, Y is explained by some low-dimensional latent
# data X, warped by a function drawn from a GP prior (f).  So Y = f(X), but
# we don't know X or f.
#
# In this example, we optimize X and the hyperparameters of the GP, but
# we integrate over all possible functions f.
#
# Normally the observed data would be high-dimensional.
#
# David Duvenaud (duvenaud@gmail.com)


import matplotlib.pyplot as plt
from data import make_pinwheel
from gaussian_process import make_gp_funs, rbf_covariance
from scipy.optimize import minimize

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import value_and_grad
from autograd.scipy.stats import norm

if __name__ == "__main__":
    data_dimension = 2  # Normally the data dimension would be much higher.
    latent_dimension = 2

    # Build model and objective function.
    params_per_gp, predict, log_marginal_likelihood = make_gp_funs(
        rbf_covariance, num_cov_params=latent_dimension + 1
    )
    total_gp_params = data_dimension * params_per_gp

    data = make_pinwheel(radial_std=0.3, tangential_std=0.05, num_classes=3, num_per_class=30, rate=0.4)
    datalen = data.shape[0]

    num_latent_params = datalen * latent_dimension

    def unpack_params(params):
        gp_params = np.reshape(params[:total_gp_params], (data_dimension, params_per_gp))
        latents = np.reshape(params[total_gp_params:], (datalen, latent_dimension))
        return gp_params, latents

    def objective(params):
        gp_params, latents = unpack_params(params)
        gp_likelihood = sum(
            [log_marginal_likelihood(gp_params[i], latents, data[:, i]) for i in range(data_dimension)]
        )
        latent_prior_likelihood = np.sum(norm.logpdf(latents))
        return -gp_likelihood - latent_prior_likelihood

    # Set up figure.
    fig = plt.figure(figsize=(12, 8), facecolor="white")
    latent_ax = fig.add_subplot(121, frameon=False)
    data_ax = fig.add_subplot(122, frameon=False)
    plt.show(block=False)

    def callback(params):
        print(f"Log likelihood {-objective(params)}")
        gp_params, latents = unpack_params(params)

        data_ax.cla()
        data_ax.plot(data[:, 0], data[:, 1], "bx")
        data_ax.set_xticks([])
        data_ax.set_yticks([])
        data_ax.set_title("Observed Data")

        latent_ax.cla()
        latent_ax.plot(latents[:, 0], latents[:, 1], "kx")
        latent_ax.set_xticks([])
        latent_ax.set_yticks([])
        latent_ax.set_xlim([-2, 2])
        latent_ax.set_ylim([-2, 2])
        latent_ax.set_title("Latent coordinates")

        plt.draw()
        plt.pause(1.0 / 60.0)

    # Initialize covariance parameters
    rs = npr.RandomState(1)
    init_params = rs.randn(total_gp_params + num_latent_params) * 0.1

    print("Optimizing covariance parameters and latent variable locations...")
    minimize(value_and_grad(objective), init_params, jac=True, method="CG", callback=callback)



from pytest import raises

from autograd.extend import defvjp
from autograd.test_util import check_grads
from autograd.tracer import primitive


def test_check_vjp_1st_order_fail():
    @primitive
    def foo(x):
        return x * 2.0

    defvjp(foo, lambda ans, x: lambda g: g * 2.001)

    with raises(AssertionError, match="\\(VJP\\) check of foo failed"):
        check_grads(foo, modes=["rev"])(1.0)


def test_check_vjp_2nd_order_fail():
    @primitive
    def foo(x):
        return x * 2.0

    defvjp(foo, lambda ans, x: lambda g: bar(g) * 2)

    @primitive
    def bar(x):
        return x

    defvjp(bar, lambda ans, x: lambda g: g * 1.001)

    with raises(AssertionError, match="\\(VJP\\) check of vjp_foo failed"):
        check_grads(foo, modes=["rev"])(1.0)



"""Implements the long-short term memory character model.
This version vectorizes over multiple examples, but each string
has a fixed length."""

from os.path import dirname, join

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad
from autograd.misc.optimizers import adam
from autograd.scipy.special import logsumexp

### Helper functions #################


def sigmoid(x):
    return 0.5 * (np.tanh(x) + 1.0)  # Output ranges from 0 to 1.


def concat_and_multiply(weights, *args):
    cat_state = np.hstack(args + (np.ones((args[0].shape[0], 1)),))
    return np.dot(cat_state, weights)


### Define recurrent neural net #######


def create_rnn_params(input_size, state_size, output_size, param_scale=0.01, rs=npr.RandomState(0)):
    return {
        "init hiddens": rs.randn(1, state_size) * param_scale,
        "change": rs.randn(input_size + state_size + 1, state_size) * param_scale,
        "predict": rs.randn(state_size + 1, output_size) * param_scale,
    }


def rnn_predict(params, inputs):
    def update_rnn(input, hiddens):
        return np.tanh(concat_and_multiply(params["change"], input, hiddens))

    def hiddens_to_output_probs(hiddens):
        output = concat_and_multiply(params["predict"], hiddens)
        return output - logsumexp(output, axis=1, keepdims=True)  # Normalize log-probs.

    num_sequences = inputs.shape[1]
    hiddens = np.repeat(params["init hiddens"], num_sequences, axis=0)
    output = [hiddens_to_output_probs(hiddens)]

    for input in inputs:  # Iterate over time steps.
        hiddens = update_rnn(input, hiddens)
        output.append(hiddens_to_output_probs(hiddens))
    return output


def rnn_log_likelihood(params, inputs, targets):
    logprobs = rnn_predict(params, inputs)
    loglik = 0.0
    num_time_steps, num_examples, _ = inputs.shape
    for t in range(num_time_steps):
        loglik += np.sum(logprobs[t] * targets[t])
    return loglik / (num_time_steps * num_examples)


### Dataset setup ##################


def string_to_one_hot(string, maxchar):
    """Converts an ASCII string to a one-of-k encoding."""
    ascii = np.array([ord(c) for c in string]).T
    return np.array(ascii[:, None] == np.arange(maxchar)[None, :], dtype=int)


def one_hot_to_string(one_hot_matrix):
    return "".join([chr(np.argmax(c)) for c in one_hot_matrix])


def build_dataset(filename, sequence_length, alphabet_size, max_lines=-1):
    """Loads a text file, and turns each line into an encoded sequence."""
    with open(filename) as f:
        content = f.readlines()
    content = content[:max_lines]
    content = [line for line in content if len(line) > 2]  # Remove blank lines
    seqs = np.zeros((sequence_length, len(content), alphabet_size))
    for ix, line in enumerate(content):
        padded_line = (line + " " * sequence_length)[:sequence_length]
        seqs[:, ix, :] = string_to_one_hot(padded_line, alphabet_size)
    return seqs


if __name__ == "__main__":
    num_chars = 128

    # Learn to predict our own source code.
    text_filename = join(dirname(__file__), "rnn.py")
    train_inputs = build_dataset(text_filename, sequence_length=30, alphabet_size=num_chars, max_lines=60)

    init_params = create_rnn_params(input_size=128, output_size=128, state_size=40, param_scale=0.01)

    def print_training_prediction(weights):
        print("Training text                         Predicted text")
        logprobs = np.asarray(rnn_predict(weights, train_inputs))
        for t in range(logprobs.shape[1]):
            training_text = one_hot_to_string(train_inputs[:, t, :])
            predicted_text = one_hot_to_string(logprobs[:, t, :])
            print(training_text.replace("\n", " ") + "|" + predicted_text.replace("\n", " "))

    def training_loss(params, iter):
        return -rnn_log_likelihood(params, train_inputs, train_inputs)

    def callback(weights, iter, gradient):
        if iter % 10 == 0:
            print("Iteration", iter, "Train loss:", training_loss(weights, 0))
            print_training_prediction(weights)

    # Build gradient of loss function using autograd.
    training_loss_grad = grad(training_loss)

    print("Training RNN...")
    trained_params = adam(training_loss_grad, init_params, step_size=0.1, num_iters=1000, callback=callback)

    print()
    print("Generating text from RNN...")
    num_letters = 30
    for t in range(20):
        text = ""
        for i in range(num_letters):
            seqs = string_to_one_hot(text, num_chars)[:, np.newaxis, :]
            logprobs = rnn_predict(trained_params, seqs)[-1].ravel()
            text += chr(npr.choice(len(logprobs), p=np.exp(logprobs)))
        print(text)



# Implements a Generative Adversarial Network, from
# arxiv.org/abs/1406.2661
# but, it always collapses to generating a single image.
# Let me know if you can get it to work! - David Duvenaud

from data import load_mnist, save_images

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad
from autograd.misc import flatten

### Define geneerator, discriminator, and objective ###


def relu(x):
    return np.maximum(0, x)


def sigmoid(x):
    return 0.5 * (np.tanh(x) + 1.0)


def logsigmoid(x):
    return x - np.logaddexp(0, x)


def init_random_params(scale, layer_sizes, rs=npr.RandomState(0)):
    """Build a list of (weights, biases) tuples,
    one for each layer in the net."""
    return [
        (
            scale * rs.randn(m, n),  # weight matrix
            scale * rs.randn(n),
        )  # bias vector
        for m, n in zip(layer_sizes[:-1], layer_sizes[1:])
    ]


def batch_normalize(activations):
    mbmean = np.mean(activations, axis=0, keepdims=True)
    return (activations - mbmean) / (np.std(activations, axis=0, keepdims=True) + 1)


def neural_net_predict(params, inputs):
    """Params is a list of (weights, bias) tuples.
    inputs is an (N x D) matrix."""
    inpW, inpb = params[0]
    inputs = relu(np.dot(inputs, inpW) + inpb)
    for W, b in params[1:-1]:
        outputs = batch_normalize(np.dot(inputs, W) + b)
        inputs = relu(outputs)
    outW, outb = params[-1]
    outputs = np.dot(inputs, outW) + outb
    return outputs


def generate_from_noise(gen_params, num_samples, noise_dim, rs):
    noise = rs.rand(num_samples, noise_dim)
    samples = neural_net_predict(gen_params, noise)
    return sigmoid(samples)


def gan_objective(gen_params, dsc_params, real_data, num_samples, noise_dim, rs):
    fake_data = generate_from_noise(gen_params, num_samples, noise_dim, rs)
    logprobs_fake = logsigmoid(neural_net_predict(dsc_params, fake_data))
    logprobs_real = logsigmoid(neural_net_predict(dsc_params, real_data))
    return np.mean(logprobs_real) - np.mean(logprobs_fake)


### Define minimax version of adam optimizer ###


def adam_minimax(
    grad_both,
    init_params_max,
    init_params_min,
    callback=None,
    num_iters=100,
    step_size_max=0.001,
    step_size_min=0.001,
    b1=0.9,
    b2=0.999,
    eps=10**-8,
):
    """Adam modified to do minimiax optimization, for instance to help with
    training generative adversarial networks."""

    x_max, unflatten_max = flatten(init_params_max)
    x_min, unflatten_min = flatten(init_params_min)

    m_max = np.zeros(len(x_max))
    v_max = np.zeros(len(x_max))
    m_min = np.zeros(len(x_min))
    v_min = np.zeros(len(x_min))
    for i in range(num_iters):
        g_max_uf, g_min_uf = grad_both(unflatten_max(x_max), unflatten_min(x_min), i)
        g_max, _ = flatten(g_max_uf)
        g_min, _ = flatten(g_min_uf)

        if callback:
            callback(
                unflatten_max(x_max), unflatten_min(x_min), i, unflatten_max(g_max), unflatten_min(g_min)
            )

        m_max = (1 - b1) * g_max + b1 * m_max  # First  moment estimate.
        v_max = (1 - b2) * (g_max**2) + b2 * v_max  # Second moment estimate.
        mhat_max = m_max / (1 - b1 ** (i + 1))  # Bias correction.
        vhat_max = v_max / (1 - b2 ** (i + 1))
        x_max = x_max + step_size_max * mhat_max / (np.sqrt(vhat_max) + eps)

        m_min = (1 - b1) * g_min + b1 * m_min  # First  moment estimate.
        v_min = (1 - b2) * (g_min**2) + b2 * v_min  # Second moment estimate.
        mhat_min = m_min / (1 - b1 ** (i + 1))  # Bias correction.
        vhat_min = v_min / (1 - b2 ** (i + 1))
        x_min = x_min - step_size_min * mhat_min / (np.sqrt(vhat_min) + eps)
    return unflatten_max(x_max), unflatten_min(x_min)


### Setup and run on MNIST ###

if __name__ == "__main__":
    # Model hyper-parameters
    noise_dim = 10
    gen_layer_sizes = [noise_dim, 200, 784]
    dsc_layer_sizes = [784, 200, 1]

    # Training parameters
    param_scale = 0.001
    batch_size = 100
    num_epochs = 50
    step_size_max = 0.01
    step_size_min = 0.01

    print("Loading training data...")
    N, train_images, _, test_images, _ = load_mnist()

    init_gen_params = init_random_params(param_scale, gen_layer_sizes)
    init_dsc_params = init_random_params(param_scale, dsc_layer_sizes)

    num_batches = int(np.ceil(len(train_images) / batch_size))

    def batch_indices(iter):
        idx = iter % num_batches
        return slice(idx * batch_size, (idx + 1) * batch_size)

    # Define training objective
    seed = npr.RandomState(0)

    def objective(gen_params, dsc_params, iter):
        idx = batch_indices(iter)
        return gan_objective(gen_params, dsc_params, train_images[idx], batch_size, noise_dim, seed)

    # Get gradients of objective using autograd.
    both_objective_grad = grad(objective, argnum=(0, 1))

    print("     Epoch     |    Objective  |       Fake probability | Real Probability  ")

    def print_perf(gen_params, dsc_params, iter, gen_gradient, dsc_gradient):
        if iter % 10 == 0:
            ability = np.mean(objective(gen_params, dsc_params, iter))
            fake_data = generate_from_noise(gen_params, 20, noise_dim, seed)
            real_data = train_images[batch_indices(iter)]
            probs_fake = np.mean(sigmoid(neural_net_predict(dsc_params, fake_data)))
            probs_real = np.mean(sigmoid(neural_net_predict(dsc_params, real_data)))
            print(f"{iter // num_batches:15}|{ability:20}|{probs_fake:20}|{probs_real:20}")
            save_images(fake_data, "gan_samples.png", vmin=0, vmax=1)

    # The optimizers provided can optimize lists, tuples, or dicts of parameters.
    optimized_params = adam_minimax(
        both_objective_grad,
        init_gen_params,
        init_dsc_params,
        step_size_max=step_size_max,
        step_size_min=step_size_min,
        num_iters=num_epochs * num_batches,
        callback=print_perf,
    )



import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad
from autograd.test_util import check_grads

npr.seed(1)


def test_real_type():
    fun = lambda x: np.sum(np.real(x))
    df = grad(fun)
    assert np.isrealobj(df(2.0))
    assert np.iscomplexobj(df(1.0j))


def test_real_if_close_type():
    fun = lambda x: np.sum(np.real(x))
    df = grad(fun)
    assert np.isrealobj(df(1.0))
    assert np.iscomplexobj(df(1.0j))


def test_angle_real():
    fun = lambda x: np.angle(x)
    d_fun = lambda x: grad(fun)(x)
    check_grads(fun)(npr.rand())
    check_grads(d_fun)(npr.rand())


def test_angle_complex():
    fun = lambda x: np.angle(x)
    d_fun = lambda x: grad(fun)(x)
    check_grads(fun)(npr.rand() + 1j * npr.rand())
    check_grads(d_fun)(npr.rand() + 1j * npr.rand())


def test_abs_real():
    fun = lambda x: np.abs(x)
    d_fun = lambda x: grad(fun)(x)
    check_grads(fun)(1.1)
    check_grads(d_fun)(2.1)


def test_abs_complex():
    fun = lambda x: np.abs(x)
    d_fun = lambda x: grad(fun)(x)
    check_grads(fun)(1.1 + 1.2j)
    check_grads(d_fun)(1.1 + 1.3j)



"""Demonstrates how to use the tracer module, independent of autodiff, by
creating a trace that prints out functions and their arguments as they're being
evaluated"""

import autograd.numpy as np  # autograd has already wrapped numpy for us
from autograd.tracer import Node, trace


class PrintNode(Node):
    def __init__(self, value, fun, args, kwargs, parent_argnums, parents):
        self.varname_generator = parents[0].varname_generator
        self.varname = next(self.varname_generator)
        args_or_vars = list(args)
        for argnum, parent in zip(parent_argnums, parents):
            args_or_vars[argnum] = parent.varname
        print("{} = {}({}) = {}".format(self.varname, fun.__name__, ",".join(map(str, args_or_vars)), value))

    def initialize_root(self, x):
        self.varname_generator = make_varname_generator()
        self.varname = next(self.varname_generator)
        print(f"{self.varname} = {x}")


def make_varname_generator():
    for i in range(65, 91):
        yield chr(i)
    raise Exception("Ran out of alphabet!")


def print_trace(f, x):
    start_node = PrintNode.new_root(x)
    trace(start_node, f, x)
    print()


def avg(x, y):
    return (x + y) / 2


def fun(x):
    y = np.sin(x + x)
    return avg(y, y)


print_trace(fun, 1.23)

# Traces can be nested, so we can also trace through grad(fun)
from autograd import grad

print_trace(grad(fun), 1.0)



import scipy.stats

import autograd.numpy as np
from autograd.extend import defvjp, primitive
from autograd.numpy.numpy_vjps import unbroadcast_f
from autograd.scipy.special import gamma

cdf = primitive(scipy.stats.chi2.cdf)
logpdf = primitive(scipy.stats.chi2.logpdf)
pdf = primitive(scipy.stats.chi2.pdf)


def grad_chi2_logpdf(x, df):
    return np.where(df % 1 == 0, (df - x - 2) / (2 * x), 0)


defvjp(
    cdf,
    lambda ans, x, df: unbroadcast_f(
        x, lambda g: g * np.power(2.0, -df / 2) * np.exp(-x / 2) * np.power(x, df / 2 - 1) / gamma(df / 2)
    ),
    argnums=[0],
)
defvjp(logpdf, lambda ans, x, df: unbroadcast_f(x, lambda g: g * grad_chi2_logpdf(x, df)), argnums=[0])
defvjp(pdf, lambda ans, x, df: unbroadcast_f(x, lambda g: g * ans * grad_chi2_logpdf(x, df)), argnums=[0])



import scipy.special

import autograd.numpy as np
from autograd.extend import defjvp, defvjp, primitive
from autograd.numpy.numpy_vjps import repeat_to_match_shape, unbroadcast_f

### Beta function ###
beta = primitive(scipy.special.beta)
betainc = primitive(scipy.special.betainc)
betaln = primitive(scipy.special.betaln)

defvjp(
    beta,
    lambda ans, a, b: unbroadcast_f(a, lambda g: g * ans * (psi(a) - psi(a + b))),
    lambda ans, a, b: unbroadcast_f(b, lambda g: g * ans * (psi(b) - psi(a + b))),
)
defvjp(
    betainc,
    lambda ans, a, b, x: unbroadcast_f(
        x, lambda g: g * np.power(x, a - 1) * np.power(1 - x, b - 1) / beta(a, b)
    ),
    argnums=[2],
)
defvjp(
    betaln,
    lambda ans, a, b: unbroadcast_f(a, lambda g: g * (psi(a) - psi(a + b))),
    lambda ans, a, b: unbroadcast_f(b, lambda g: g * (psi(b) - psi(a + b))),
)

### Gamma functions ###
polygamma = primitive(scipy.special.polygamma)
psi = primitive(scipy.special.psi)  # psi(x) is just polygamma(0, x)
digamma = primitive(scipy.special.digamma)  # digamma is another name for psi.
gamma = primitive(scipy.special.gamma)
gammaln = primitive(scipy.special.gammaln)
gammainc = primitive(scipy.special.gammainc)
gammaincc = primitive(scipy.special.gammaincc)
gammasgn = primitive(scipy.special.gammasgn)
rgamma = primitive(scipy.special.rgamma)
multigammaln = primitive(scipy.special.multigammaln)

defvjp(gammasgn, None)
defvjp(polygamma, None, lambda ans, n, x: lambda g: g * polygamma(n + 1, x))
defvjp(psi, lambda ans, x: lambda g: g * polygamma(1, x))
defvjp(digamma, lambda ans, x: lambda g: g * polygamma(1, x))
defvjp(gamma, lambda ans, x: lambda g: g * ans * psi(x))
defvjp(gammaln, lambda ans, x: lambda g: g * psi(x))
defvjp(rgamma, lambda ans, x: lambda g: g * psi(x) / -gamma(x))
defvjp(
    multigammaln,
    lambda ans, a, d: lambda g: g * np.sum(digamma(np.expand_dims(a, -1) - np.arange(d) / 2.0), -1),
    None,
)


def make_gammainc_vjp_arg1(sign):
    def gammainc_vjp_arg1(ans, a, x):
        coeffs = sign * np.exp(-x) * np.power(x, a - 1) / gamma(a)
        return unbroadcast_f(x, lambda g: g * coeffs)

    return gammainc_vjp_arg1


defvjp(gammainc, make_gammainc_vjp_arg1(1), argnums=[1])
defvjp(gammaincc, make_gammainc_vjp_arg1(-1), argnums=[1])

### Bessel functions ###

j0 = primitive(scipy.special.j0)
y0 = primitive(scipy.special.y0)
j1 = primitive(scipy.special.j1)
y1 = primitive(scipy.special.y1)
jn = primitive(scipy.special.jn)
yn = primitive(scipy.special.yn)

defvjp(j0, lambda ans, x: lambda g: -g * j1(x))
defvjp(y0, lambda ans, x: lambda g: -g * y1(x))
defvjp(j1, lambda ans, x: lambda g: g * (j0(x) - jn(2, x)) / 2.0)
defvjp(y1, lambda ans, x: lambda g: g * (y0(x) - yn(2, x)) / 2.0)
defvjp(jn, None, lambda ans, n, x: lambda g: g * (jn(n - 1, x) - jn(n + 1, x)) / 2.0)
defvjp(yn, None, lambda ans, n, x: lambda g: g * (yn(n - 1, x) - yn(n + 1, x)) / 2.0)


### Faster versions of common Bessel functions ###
i0 = primitive(scipy.special.i0)
i1 = primitive(scipy.special.i1)
iv = primitive(scipy.special.iv)
ive = primitive(scipy.special.ive)

defvjp(i0, lambda ans, x: lambda g: g * i1(x))
defvjp(i1, lambda ans, x: lambda g: g * (i0(x) + iv(2, x)) / 2.0)
defvjp(iv, None, lambda ans, n, x: lambda g: g * (iv(n - 1, x) + iv(n + 1, x)) / 2.0)
defvjp(ive, None, lambda ans, n, x: lambda g: g * (ans * (n / x - np.sign(x)) + ive(n + 1, x)))

### Error Function ###
inv_root_pi = 0.56418958354775627928
erf = primitive(scipy.special.erf)
erfc = primitive(scipy.special.erfc)

defvjp(erf, lambda ans, x: lambda g: 2.0 * g * inv_root_pi * np.exp(-(x**2)))
defvjp(erfc, lambda ans, x: lambda g: -2.0 * g * inv_root_pi * np.exp(-(x**2)))


### Inverse error function ###
root_pi = 1.7724538509055159
erfinv = primitive(scipy.special.erfinv)
erfcinv = primitive(scipy.special.erfcinv)

defvjp(erfinv, lambda ans, x: lambda g: g * root_pi / 2 * np.exp(erfinv(x) ** 2))
defvjp(erfcinv, lambda ans, x: lambda g: -g * root_pi / 2 * np.exp(erfcinv(x) ** 2))

### Logit and Expit ###
logit = primitive(scipy.special.logit)
expit = primitive(scipy.special.expit)

defvjp(logit, lambda ans, x: lambda g: g / (x * (1 - x)))
defvjp(expit, lambda ans, x: lambda g: g * ans * (1 - ans))

### logsumexp ###
logsumexp = primitive(scipy.special.logsumexp)


def make_grad_logsumexp(ans, x, axis=None, b=1.0, keepdims=False):
    shape, dtype = np.shape(x), np.result_type(x)

    def vjp(g):
        g_repeated, _ = repeat_to_match_shape(g, shape, dtype, axis, keepdims)
        ans_repeated, _ = repeat_to_match_shape(ans, shape, dtype, axis, keepdims)
        return g_repeated * b * np.exp(x - ans_repeated)

    return vjp


defvjp(logsumexp, make_grad_logsumexp)


def fwd_grad_logsumexp(g, ans, x, axis=None, b=1.0, keepdims=False):
    if not keepdims:
        if isinstance(axis, int):
            ans = np.expand_dims(ans, axis)
        elif isinstance(axis, tuple):
            for ax in sorted(axis):
                ans = np.expand_dims(ans, ax)
    return np.sum(g * b * np.exp(x - ans), axis=axis, keepdims=keepdims)


defjvp(logsumexp, fwd_grad_logsumexp)



"""Generates a graphviz DOT file of an evaluation trace.
Usage (need the dot binary, from the graphviz package, www.graphviz.org):

python2 dot_graph.py | dot -Tpdf -o graph.pdf
"""

import autograd.numpy as np
from autograd.tracer import Node, trace


class GraphNode(Node):
    # Records the full graph (could having this in tracer.py)
    def __init__(self, value, fun, args, kwargs, parent_argnums, parents):
        self.fun_name = fun.__name__
        self.args = args
        self.parents = dict(zip(parent_argnums, parents))
        self.isroot = False

    def initialize_root(self, x):
        self.isroot = True

    def __repr__(self):
        return f"node_{id(self)}"


def trace_graph(f, x):
    start_node = GraphNode.new_root(x)
    _, node = trace(start_node, f, x)
    return node


dot_edge = "{} -> {} [color=gray30];\n".format
dot_function_node = '{} [label="{}", shape=box, color=lightblue, style=filled];\n'.format
dot_variable_node = '{} [label="{}", color=orange, style=filled];\n'.format
dot_graph = "digraph G {{{}}}".format


def graph_to_dotfile(graph):
    visited = set()

    def node_to_fragment(node):
        visited.add(node)
        if node.isroot:
            return dot_variable_node(node, "input")
        fragment = dot_function_node(node, node.fun_name)
        for argnum, arg in enumerate(node.args):
            if argnum in node.parents:
                parent = node.parents[argnum]
                fragment += dot_edge(parent, node)
                if parent not in visited:
                    fragment += node_to_fragment(parent)
            else:
                argnode = f"{node}_arg_{argnum}"
                fragment += dot_edge(argnode, node)
                fragment += dot_variable_node(argnode, arg)

        return fragment

    dot_body = node_to_fragment(graph)
    dot_body += dot_variable_node("output", "output")
    dot_body += dot_edge(graph, "output")
    return dot_graph(dot_body)


if __name__ == "__main__":

    def fun(x):
        y = np.sin(x)
        return (y + np.exp(x) - 0.5) * y

    print(graph_to_dotfile(trace_graph(fun, 1.0)))



import matplotlib.pyplot as plt

import autograd.numpy as np
import autograd.numpy.random as npr
import autograd.scipy.stats.norm as norm
from autograd import grad
from autograd.misc import flatten
from autograd.misc.optimizers import adam


def init_random_params(scale, layer_sizes, rs=npr.RandomState(0)):
    """Build a list of (weights, biases) tuples, one for each layer."""
    return [
        (
            rs.randn(insize, outsize) * scale,  # weight matrix
            rs.randn(outsize) * scale,
        )  # bias vector
        for insize, outsize in zip(layer_sizes[:-1], layer_sizes[1:])
    ]


def nn_predict(params, inputs, nonlinearity=np.tanh):
    for W, b in params:
        outputs = np.dot(inputs, W) + b
        inputs = nonlinearity(outputs)
    return outputs


def log_gaussian(params, scale):
    flat_params, _ = flatten(params)
    return np.sum(norm.logpdf(flat_params, 0, scale))


def logprob(weights, inputs, targets, noise_scale=0.1):
    predictions = nn_predict(weights, inputs)
    return np.sum(norm.logpdf(predictions, targets, noise_scale))


def build_toy_dataset(n_data=80, noise_std=0.1):
    rs = npr.RandomState(0)
    inputs = np.concatenate([np.linspace(0, 3, num=n_data / 2), np.linspace(6, 8, num=n_data / 2)])
    targets = np.cos(inputs) + rs.randn(n_data) * noise_std
    inputs = (inputs - 4.0) / 2.0
    inputs = inputs[:, np.newaxis]
    targets = targets[:, np.newaxis] / 2.0
    return inputs, targets


if __name__ == "__main__":
    init_scale = 0.1
    weight_prior_variance = 10.0
    init_params = init_random_params(init_scale, layer_sizes=[1, 4, 4, 1])

    inputs, targets = build_toy_dataset()

    def objective(weights, t):
        return -logprob(weights, inputs, targets) - log_gaussian(weights, weight_prior_variance)

    print(grad(objective)(init_params, 0))

    # Set up figure.
    fig = plt.figure(figsize=(12, 8), facecolor="white")
    ax = fig.add_subplot(111, frameon=False)
    plt.show(block=False)

    def callback(params, t, g):
        print(f"Iteration {t} log likelihood {-objective(params, t)}")

        # Plot data and functions.
        plt.cla()
        ax.plot(inputs.ravel(), targets.ravel(), "bx", ms=12)
        plot_inputs = np.reshape(np.linspace(-7, 7, num=300), (300, 1))
        outputs = nn_predict(params, plot_inputs)
        ax.plot(plot_inputs, outputs, "r", lw=3)
        ax.set_ylim([-1, 1])
        plt.draw()
        plt.pause(1.0 / 60.0)

    print("Optimizing network parameters...")
    optimized_params = adam(grad(objective), init_params, step_size=0.01, num_iters=1000, callback=callback)



"""This Bayesian optimization demo using gradient-based optimization
to find the next query point."""

import matplotlib.pyplot as plt
from gaussian_process import make_gp_funs, rbf_covariance
from scipy.optimize import minimize

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import value_and_grad
from autograd.scipy.stats import norm


def probability_of_improvement(mean, std, max_so_far):
    return norm.cdf(max_so_far, mean, std)


def expected_new_max(mean, std, max_so_far):
    return (
        max_so_far
        - (mean - max_so_far) * norm.cdf(mean, max_so_far, std)
        + std * norm.pdf(mean, max_so_far, std)
    )


def init_covariance_params(num_params):
    return np.zeros(num_params)


def defaultmax(x, default=-np.inf):
    if x.size == 0:
        return default
    return np.max(x)


def bayesian_optimize(func, domain_min, domain_max, num_iters=20, callback=None):
    D = len(domain_min)

    num_params, predict, log_marginal_likelihood = make_gp_funs(rbf_covariance, num_cov_params=D + 1)

    model_params = init_covariance_params(num_params)

    def optimize_gp_params(init_params, X, y):
        log_hyperprior = lambda params: np.sum(norm.logpdf(params, 0.0, 100.0))
        objective = lambda params: -log_marginal_likelihood(params, X, y) - log_hyperprior(params)
        return minimize(value_and_grad(objective), init_params, jac=True, method="CG").x

    def choose_next_point(domain_min, domain_max, acquisition_function, num_tries=15, rs=npr.RandomState(0)):
        """Uses gradient-based optimization to find next query point."""
        init_points = rs.rand(num_tries, D) * (domain_max - domain_min) + domain_min

        grad_obj = value_and_grad(lambda x: -acquisition_function(x))

        def optimize_point(init_point):
            print(".", end="")
            result = minimize(
                grad_obj,
                x0=init_point,
                jac=True,
                method="L-BFGS-B",
                options={"maxiter": 10},
                bounds=list(zip(domain_min, domain_max)),
            )
            return result.x, acquisition_function(result.x)

        optimzed_points, optimized_values = list(zip(*list(map(optimize_point, init_points))))
        print()
        best_ix = np.argmax(optimized_values)
        return np.atleast_2d(optimzed_points[best_ix])

    # Start by evaluating once in the middle of the domain.
    X = np.zeros((0, D))
    y = np.zeros(0)
    X = np.concatenate((X, np.reshape((domain_max - domain_min) / 2.0, (D, 1))))
    y = np.concatenate((y, np.reshape(np.array(func(X)), (1,))))

    for i in range(num_iters):
        if i > 1:
            print("Optimizing model parameters...")
            model_params = optimize_gp_params(model_params, X, y)

        print("Choosing where to look next", end="")

        def predict_func(xstar):
            mean, cov = predict(model_params, X, y, xstar)
            return mean, np.sqrt(np.diag(cov))

        def acquisition_function(xstar):
            xstar = np.atleast_2d(xstar)  # To work around a bug in scipy.minimize
            mean, std = predict_func(xstar)
            return expected_new_max(mean, std, defaultmax(y))

        next_point = choose_next_point(domain_min, domain_max, acquisition_function)

        print("Evaluating expensive function...")
        new_value = func(next_point)

        X = np.concatenate((X, next_point))
        y = np.concatenate((y, np.reshape(np.array(new_value), (1,))))

        if callback:
            callback(X, y, predict_func, acquisition_function, next_point, new_value)

    best_ix = np.argmax(y)
    return X[best_ix, :], y[best_ix]


if __name__ == "__main__":

    def example_function(x):
        return np.sum(x * np.sin(10.0 * x) + x) - 1

    domain_min = np.array([0.0])
    domain_max = np.array([1.1])

    # Set up figure.
    fig = plt.figure(figsize=(12, 8), facecolor="white")
    ax = fig.add_subplot(111, frameon=False)
    plt.show(block=False)

    def callback(X, y, predict_func, acquisition_function, next_point, new_value):
        plt.cla()

        # Show posterior marginals.
        plot_xs = np.reshape(np.linspace(domain_min, domain_max, 300), (300, 1))
        pred_mean, pred_std = predict_func(plot_xs)
        ax.plot(plot_xs, pred_mean, "b")
        ax.fill(
            np.concatenate([plot_xs, plot_xs[::-1]]),
            np.concatenate([pred_mean - 1.96 * pred_std, (pred_mean + 1.96 * pred_std)[::-1]]),
            alpha=0.15,
            fc="Blue",
            ec="None",
        )

        ax.plot(X, y, "kx")
        ax.plot(next_point, new_value, "ro")

        alphas = acquisition_function(plot_xs)
        ax.plot(plot_xs, alphas, "r")
        ax.set_ylim([-1.5, 1.5])
        ax.set_xticks([])
        ax.set_yticks([])
        plt.draw()
        plt.pause(1)

    best_x, best_y = bayesian_optimize(example_function, domain_min, domain_max, callback=callback)



import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad
from autograd import isinstance as ag_isinstance
from autograd import tuple as ag_tuple
from autograd.test_util import check_grads

npr.seed(1)


def test_getter():
    def fun(input_tuple):
        A = np.sum(input_tuple[0])
        B = np.sum(input_tuple[1])
        C = np.sum(input_tuple[1])
        return A + B + C

    d_fun = grad(fun)
    input_tuple = (npr.randn(5, 6), npr.randn(4, 3), npr.randn(2, 4))

    result = d_fun(input_tuple)
    assert np.allclose(result[0], np.ones((5, 6)))
    assert np.allclose(result[1], 2 * np.ones((4, 3)))
    assert np.allclose(result[2], np.zeros((2, 4)))


def test_grads():
    def fun(input_tuple):
        A = np.sum(np.sin(input_tuple[0]))
        B = np.sum(np.cos(input_tuple[1]))
        return A + B

    def d_fun(input_tuple):
        g = grad(fun)(input_tuple)
        A = np.sum(g[0])
        B = np.sum(np.sin(g[0]))
        C = np.sum(np.sin(g[1]))
        return A + B + C

    input_tuple = (npr.randn(5, 6), npr.randn(4, 3), npr.randn(2, 4))

    check_grads(fun)(input_tuple)
    check_grads(d_fun)(input_tuple)


def test_nested_higher_order():
    def outer_fun(x):
        def inner_fun(y):
            return y[0] * y[1]

        return np.sum(np.sin(np.array(grad(inner_fun)(ag_tuple((x, x))))))

    check_grads(outer_fun)(5.0)
    check_grads(grad(outer_fun))(10.0)
    check_grads(grad(grad(outer_fun)))(10.0)


def test_isinstance():
    def fun(x):
        assert ag_isinstance(x, tuple)
        assert ag_isinstance(x, ag_tuple)
        return x[0]

    fun((1.0, 2.0, 3.0))
    grad(fun)((1.0, 2.0, 3.0))



import numpy as onp

from autograd.extend import JVPNode, def_linear, defjvp, defjvp_argnum, register_notrace, vspace

from ..util import func
from . import numpy_wrapper as anp
from .numpy_boxes import ArrayBox
from .numpy_vjps import (
    balanced_eq,
    dot_adjoint_0,
    dot_adjoint_1,
    match_complex,
    nograd_functions,
    replace_zero,
    tensordot_adjoint_0,
    tensordot_adjoint_1,
    untake,
)

for fun in nograd_functions:
    register_notrace(JVPNode, fun)

defjvp(func(ArrayBox.__getitem__), "same")
defjvp(untake, "same")

defjvp_argnum(anp.array_from_args, lambda argnum, g, ans, args, kwargs: untake(g, argnum - 2, vspace(ans)))
defjvp(
    anp._array_from_scalar_or_array,
    None,
    None,
    lambda g, ans, args, kwargs, _: anp._array_from_scalar_or_array(args, kwargs, g),
)

# ----- Functions that are constant w.r.t. continuous inputs -----
defjvp(anp.nan_to_num, lambda g, ans, x: anp.where(anp.isfinite(x), g, 0.0))

# ----- Binary ufuncs (linear) -----
def_linear(anp.multiply)

# ----- Binary ufuncs -----
defjvp(anp.add, lambda g, ans, x, y: broadcast(g, ans), lambda g, ans, x, y: broadcast(g, ans))
defjvp(anp.subtract, lambda g, ans, x, y: broadcast(g, ans), lambda g, ans, x, y: broadcast(-g, ans))
defjvp(anp.divide, "same", lambda g, ans, x, y: -g * x / y**2)
defjvp(
    anp.maximum,
    lambda g, ans, x, y: g * balanced_eq(x, ans, y),
    lambda g, ans, x, y: g * balanced_eq(y, ans, x),
)
defjvp(
    anp.minimum,
    lambda g, ans, x, y: g * balanced_eq(x, ans, y),
    lambda g, ans, x, y: g * balanced_eq(y, ans, x),
)
defjvp(
    anp.fmax,
    lambda g, ans, x, y: g * balanced_eq(x, ans, y),
    lambda g, ans, x, y: g * balanced_eq(y, ans, x),
)
defjvp(
    anp.fmin,
    lambda g, ans, x, y: g * balanced_eq(x, ans, y),
    lambda g, ans, x, y: g * balanced_eq(y, ans, x),
)
defjvp(anp.logaddexp, lambda g, ans, x, y: g * anp.exp(x - ans), lambda g, ans, x, y: g * anp.exp(y - ans))
defjvp(anp.logaddexp2, lambda g, ans, x, y: g * 2 ** (x - ans), lambda g, ans, x, y: g * 2 ** (y - ans))
defjvp(anp.true_divide, "same", lambda g, ans, x, y: -g * x / y**2)
defjvp(anp.mod, lambda g, ans, x, y: broadcast(g, ans), lambda g, ans, x, y: -g * anp.floor(x / y))
defjvp(anp.remainder, lambda g, ans, x, y: broadcast(g, ans), lambda g, ans, x, y: -g * anp.floor(x / y))
defjvp(
    anp.power,
    lambda g, ans, x, y: g * y * x ** anp.where(y, y - 1, 1.0),
    lambda g, ans, x, y: g * anp.log(replace_zero(x, 1.0)) * ans,
)
defjvp(anp.arctan2, lambda g, ans, x, y: g * y / (x**2 + y**2), lambda g, ans, x, y: g * -x / (x**2 + y**2))

# ----- Simple grads (linear) -----
defjvp(anp.negative, "same")
defjvp(anp.rad2deg, "same")
defjvp(anp.degrees, "same")
defjvp(anp.deg2rad, "same")
defjvp(anp.radians, "same")
defjvp(anp.reshape, "same")
defjvp(anp.roll, "same")
defjvp(anp.array_split, "same")
defjvp(anp.split, "same")
defjvp(anp.vsplit, "same")
defjvp(anp.hsplit, "same")
defjvp(anp.dsplit, "same")
defjvp(anp.ravel, "same")
defjvp(anp.expand_dims, "same")
defjvp(anp.squeeze, "same")
defjvp(anp.diag, "same")
defjvp(anp.diagonal, "same")
defjvp(anp.make_diagonal, "same")
defjvp(anp.flipud, "same")
defjvp(anp.fliplr, "same")
defjvp(anp.rot90, "same")
defjvp(anp.trace, "same")
defjvp(anp.full, "same", argnums=(1,))
defjvp(anp.triu, "same")
defjvp(anp.tril, "same")
defjvp(anp.swapaxes, "same")
defjvp(anp.rollaxis, "same")
defjvp(anp.moveaxis, "same")
defjvp(anp.broadcast_to, "same")
def_linear(anp.cross)

# ----- Simple grads -----
defjvp(anp.abs, lambda g, ans, x: anp.real(g * replace_zero(anp.conj(x), 0.0)) / replace_zero(ans, 1.0))
defjvp(anp.fabs, lambda g, ans, x: anp.sign(x) * g)  # fabs doesn't take complex numbers.
defjvp(anp.absolute, lambda g, ans, x: anp.real(g * anp.conj(x)) / ans)
defjvp(anp.reciprocal, lambda g, ans, x: -g / x**2)
defjvp(anp.exp, lambda g, ans, x: ans * g)
defjvp(anp.exp2, lambda g, ans, x: ans * anp.log(2) * g)
defjvp(anp.expm1, lambda g, ans, x: (ans + 1) * g)
defjvp(anp.log, lambda g, ans, x: g / x)
defjvp(anp.log2, lambda g, ans, x: g / x / anp.log(2))
defjvp(anp.log10, lambda g, ans, x: g / x / anp.log(10))
defjvp(anp.log1p, lambda g, ans, x: g / (x + 1))
defjvp(anp.sin, lambda g, ans, x: g * anp.cos(x))
defjvp(anp.cos, lambda g, ans, x: -g * anp.sin(x))
defjvp(anp.tan, lambda g, ans, x: g / anp.cos(x) ** 2)
defjvp(anp.arcsin, lambda g, ans, x: g / anp.sqrt(1 - x**2))
defjvp(anp.arccos, lambda g, ans, x: -g / anp.sqrt(1 - x**2))
defjvp(anp.arctan, lambda g, ans, x: g / (1 + x**2))
defjvp(anp.sinh, lambda g, ans, x: g * anp.cosh(x))
defjvp(anp.cosh, lambda g, ans, x: g * anp.sinh(x))
defjvp(anp.tanh, lambda g, ans, x: g / anp.cosh(x) ** 2)
defjvp(anp.arcsinh, lambda g, ans, x: g / anp.sqrt(x**2 + 1))
defjvp(anp.arccosh, lambda g, ans, x: g / anp.sqrt(x**2 - 1))
defjvp(anp.arctanh, lambda g, ans, x: g / (1 - x**2))
defjvp(anp.square, lambda g, ans, x: g * 2 * x)
defjvp(anp.sqrt, lambda g, ans, x: g * 0.5 * x**-0.5)
defjvp(
    anp.sinc,
    lambda g, ans, x: g * (anp.cos(anp.pi * x) * anp.pi * x - anp.sin(anp.pi * x)) / (anp.pi * x**2),
)
defjvp(anp.clip, lambda g, ans, x, a_min, a_max: g * anp.logical_and(ans != a_min, ans != a_max))
defjvp(anp.real_if_close, lambda g, ans, x: match_complex(ans, g))
defjvp(anp.real, lambda g, ans, x: anp.real(g))
defjvp(anp.imag, lambda g, ans, x: match_complex(ans, -1j * g))
defjvp(anp.conj, lambda g, ans, x: anp.conj(g))
defjvp(anp.angle, lambda g, ans, x: match_complex(ans, g * anp.conj(x * 1j) / anp.abs(x) ** 2))
defjvp(
    anp.where,
    None,
    lambda g, ans, c, x=None, y=None: anp.where(c, g, anp.zeros(anp.shape(g))),
    lambda g, ans, c, x=None, y=None: anp.where(c, anp.zeros(g.shape), g),
)

# ----- Trickier grads -----
defjvp(anp.kron, "same", "same")
defjvp(anp.diff, "same")
defjvp(anp.gradient, "same")
defjvp(anp.repeat, "same")
defjvp(anp.tile, "same")
defjvp(anp.transpose, "same")
defjvp(anp.sum, "same")
defjvp(anp.mean, "same")
defjvp(
    anp.prod, lambda g, ans, x, axis=None, keepdims=False: ans * anp.sum(g / x, axis=axis, keepdims=keepdims)
)
defjvp(
    anp.linspace,
    lambda g, ans, start, stop, *args, **kwargs: anp.linspace(g, 0, *args, **kwargs),
    lambda g, ans, start, stop, *args, **kwargs: anp.linspace(0, g, *args, **kwargs),
)


def forward_grad_np_var(g, ans, x, axis=None, ddof=0, keepdims=False):
    if axis is None:
        num_reps = anp.size(g)
    elif isinstance(axis, int):
        num_reps = anp.shape(g)[axis]
    elif isinstance(axis, tuple):
        num_reps = anp.prod(anp.array(np.shape(g))[list(axis)])

    x_minus_mean = anp.conj(x - anp.mean(x, axis=axis, keepdims=True))
    return 2.0 * anp.sum(anp.real(g * x_minus_mean), axis=axis, keepdims=keepdims) / (num_reps - ddof)


defjvp(anp.var, forward_grad_np_var)


def forward_grad_np_std(g, ans, x, axis=None, ddof=0, keepdims=False):
    if axis is None:
        num_reps = anp.size(g)
    elif isinstance(axis, int):
        num_reps = anp.shape(g)[axis]
    elif isinstance(axis, tuple):
        num_reps = anp.prod(anp.array(anp.shape(g))[list(axis)])

    if num_reps <= 1:
        return anp.zeros_like(ans)
    x_minus_mean = anp.conj(x - anp.mean(x, axis=axis, keepdims=True))
    return anp.sum(anp.real(g * x_minus_mean), axis=axis, keepdims=keepdims) / ((num_reps - ddof) * ans)


defjvp(anp.std, forward_grad_np_std)


def fwd_grad_chooser(g, ans, x, axis=None, keepdims=False):
    if anp.isscalar(x):
        return g
    if not keepdims:
        if isinstance(axis, int):
            ans = anp.expand_dims(ans, axis)
        elif isinstance(axis, tuple):
            for ax in sorted(axis):
                ans = anp.expand_dims(ans, ax)
    chosen_locations = x == ans
    return anp.sum((g * chosen_locations), axis=axis, keepdims=keepdims) / anp.sum(
        chosen_locations, axis=axis, keepdims=keepdims
    )


defjvp(anp.max, fwd_grad_chooser)
defjvp(anp.min, fwd_grad_chooser)
defjvp(anp.amax, fwd_grad_chooser)
defjvp(anp.amin, fwd_grad_chooser)

defjvp(anp.cumsum, "same")

def_linear(anp.inner)
def_linear(anp.matmul)
def_linear(anp.dot)
def_linear(anp.tensordot)
def_linear(anp.outer)

def_linear(dot_adjoint_0)
def_linear(dot_adjoint_1)

def_linear(tensordot_adjoint_0)
def_linear(tensordot_adjoint_1)


def fwd_grad_concatenate_args(argnum, g, ans, axis_args, kwargs):
    result = []
    for i in range(1, len(axis_args)):
        if i == argnum:
            result.append(g)
        else:
            result.append(anp.zeros_like(axis_args[i]))
    return anp.concatenate_args(axis_args[0], *result)


defjvp_argnum(anp.concatenate_args, fwd_grad_concatenate_args)


def fwd_grad_sort(g, ans, x, axis=-1, kind="quicksort", order=None):
    sort_perm = anp.argsort(x, axis, kind, order)
    return g[sort_perm]


defjvp(anp.sort, fwd_grad_sort)
if onp.lib.NumpyVersion(onp.__version__) < "2.0.0":
    defjvp(anp.msort, lambda g, ans, x: fwd_grad_sort(g, ans, x, axis=0))


def fwd_grad_partition(g, ans, x, kth, axis=-1, kind="introselect", order=None):
    partition_perm = anp.argpartition(x, kth, axis, kind, order)
    return g[partition_perm]


defjvp(anp.partition, fwd_grad_partition)


def atleast_jvpmaker(fun):
    def jvp(g, ans, *arys):
        if len(arys) > 1:
            raise NotImplementedError("Can't handle multiple arguments yet.")
        return fun(g)

    return jvp


defjvp(anp.atleast_1d, atleast_jvpmaker(anp.atleast_1d))
defjvp(anp.atleast_2d, atleast_jvpmaker(anp.atleast_2d))
defjvp(anp.atleast_3d, atleast_jvpmaker(anp.atleast_3d))

def_linear(anp.einsum)


# TODO(mattjj): can we call np.broadcast_to or a related function instead?
def broadcast(x, target):
    target_shape, target_ndim, target_dtype, target_iscomplex = anp.metadata(target)
    while anp.ndim(x) < target_ndim:
        x = anp.expand_dims(x, 0)
    for axis, size in enumerate(anp.shape(x)):
        if size == 1:
            x = anp.repeat(x, target_shape[axis], axis=axis)
    if target_iscomplex and not anp.iscomplexobj(x):
        x = x + 0j  # TODO(mattjj): this might promote the dtype
    return x


defjvp(anp.pad, lambda g, ans, array, width, mode, **kwargs: anp.pad(g, width, mode))



import matplotlib.pyplot as plt

import autograd.numpy as np
from autograd import elementwise_grad as egrad

"""
Mathematically we can only take gradients of scalar-valued functions, but
autograd's elementwise_grad function also handles numpy's familiar vectorization
of scalar functions, which is used in this example.

To be precise, elementwise_grad(fun)(x) always returns the value of a
vector-Jacobian product, where the Jacobian of fun is evaluated at x and the
vector is an all-ones vector with the same size as the output of fun. When
vectorizing a scalar-valued function over many arguments, the Jacobian of the
overall vector-to-vector mapping is diagonal, and so this vector-Jacobian
product simply returns the diagonal elements of the Jacobian, which is the
(elementwise) gradient of the function at each input value over which the
function is vectorized.
"""


def tanh(x):
    return (1.0 - np.exp((-2 * x))) / (1.0 + np.exp(-(2 * x)))


### Plotting
plt.figure(figsize=(12, 8))
x = np.linspace(-7, 7, 700)
plt.plot(x, tanh(x), label="tanh(x)")
plt.plot(x, egrad(tanh)(x), label="1st derivative")
plt.plot(x, egrad(egrad(tanh))(x), label="2nd derivative")
plt.plot(x, egrad(egrad(egrad(tanh)))(x), label="3rd derivative")
plt.plot(x, egrad(egrad(egrad(egrad(tanh))))(x), label="4th derivative")
plt.xlabel("x")
plt.ylabel("y")
plt.ylim(-5, 5)
plt.yticks(np.arange(-5, 6, 1))
plt.legend()
plt.grid(True)
plt.title("tanh(x) and its derivatives")
plt.savefig("tanh.png")
plt.show()



from functools import partial

import numpy as npo  # original numpy
from numpy.lib.stride_tricks import as_strided

import autograd.numpy as np
from autograd.extend import defvjp, primitive


@primitive
def convolve(A, B, axes=None, dot_axes=[(), ()], mode="full"):
    assert mode in ["valid", "full"], f"Mode {mode} not yet implemented"
    if axes is None:
        axes = [list(range(A.ndim)), list(range(A.ndim))]
    wrong_order = any([B.shape[ax_B] < A.shape[ax_A] for ax_A, ax_B in zip(*axes)])
    if wrong_order:
        if mode == "valid" and not all([B.shape[ax_B] <= A.shape[ax_A] for ax_A, ax_B in zip(*axes)]):
            raise Exception("One array must be larger than the other along all convolved dimensions")
        elif mode != "full" or B.size <= A.size:  # Tie breaker
            i1 = B.ndim - len(dot_axes[1]) - len(axes[1])  # B ignore
            i2 = i1 + A.ndim - len(dot_axes[0]) - len(axes[0])  # A ignore
            i3 = i2 + len(axes[0])
            ignore_B = list(range(i1))
            ignore_A = list(range(i1, i2))
            conv = list(range(i2, i3))
            return convolve(B, A, axes=axes[::-1], dot_axes=dot_axes[::-1], mode=mode).transpose(
                ignore_A + ignore_B + conv
            )

    if mode == "full":
        B = pad_to_full(B, A, axes[::-1])
    B_view_shape = list(B.shape)
    B_view_strides = list(B.strides)
    flipped_idxs = [slice(None)] * A.ndim
    for ax_A, ax_B in zip(*axes):
        B_view_shape.append(abs(B.shape[ax_B] - A.shape[ax_A]) + 1)
        B_view_strides.append(B.strides[ax_B])
        B_view_shape[ax_B] = A.shape[ax_A]
        flipped_idxs[ax_A] = slice(None, None, -1)
    B_view = as_strided(B, B_view_shape, B_view_strides)
    A_view = A[tuple(flipped_idxs)]
    all_axes = [list(axes[i]) + list(dot_axes[i]) for i in [0, 1]]
    return einsum_tensordot(A_view, B_view, all_axes)


def einsum_tensordot(A, B, axes, reverse=False):
    # Does tensor dot product using einsum, which shouldn't require a copy.
    A_axnums = list(range(A.ndim))
    B_axnums = list(range(A.ndim, A.ndim + B.ndim))
    sum_axnum = A.ndim + B.ndim
    for i_sum, (i_A, i_B) in enumerate(zip(*axes)):
        A_axnums[i_A] = sum_axnum + i_sum
        B_axnums[i_B] = sum_axnum + i_sum
    return npo.einsum(A, A_axnums, B, B_axnums)


def pad_to_full(A, B, axes):
    A_pad = [(0, 0)] * A.ndim
    for ax_A, ax_B in zip(*axes):
        A_pad[ax_A] = (B.shape[ax_B] - 1,) * 2
    return npo.pad(A, A_pad, mode="constant")


def parse_axes(A_shape, B_shape, conv_axes, dot_axes, mode):
    A_ndim, B_ndim = len(A_shape), len(B_shape)
    if conv_axes is None:
        conv_axes = (
            tuple(range(A_ndim)),
            tuple(range(A_ndim)),
        )
    axes = {
        "A": {
            "conv": tuple(conv_axes[0]),
            "dot": tuple(dot_axes[0]),
            "ignore": tuple(i for i in range(A_ndim) if i not in conv_axes[0] and i not in dot_axes[0]),
        },
        "B": {
            "conv": tuple(conv_axes[1]),
            "dot": tuple(dot_axes[1]),
            "ignore": tuple(i for i in range(B_ndim) if i not in conv_axes[1] and i not in dot_axes[1]),
        },
    }
    assert len(axes["A"]["dot"]) == len(axes["B"]["dot"])
    assert len(axes["A"]["conv"]) == len(axes["B"]["conv"])
    i1 = len(axes["A"]["ignore"])
    i2 = i1 + len(axes["B"]["ignore"])
    i3 = i2 + len(axes["A"]["conv"])
    axes["out"] = {
        "ignore_A": tuple(range(i1)),
        "ignore_B": tuple(range(i1, i2)),
        "conv": tuple(range(i2, i3)),
    }
    conv_shape = (
        compute_conv_size(A_shape[i], B_shape[j], mode) for i, j in zip(axes["A"]["conv"], axes["B"]["conv"])
    )
    shapes = {
        "A": {s: tuple(A_shape[i] for i in ax) for s, ax in axes["A"].items()},
        "B": {s: tuple(B_shape[i] for i in ax) for s, ax in axes["B"].items()},
    }
    shapes["out"] = {
        "ignore_A": shapes["A"]["ignore"],
        "ignore_B": shapes["B"]["ignore"],
        "conv": conv_shape,
    }
    return axes, shapes


def compute_conv_size(A_size, B_size, mode):
    if mode == "full":
        return A_size + B_size - 1
    elif mode == "same":
        return A_size
    elif mode == "valid":
        return abs(A_size - B_size) + 1
    else:
        raise Exception(f"Mode {mode} not recognized")


def flipped_idxs(ndim, axes):
    new_idxs = [slice(None)] * ndim
    for ax in axes:
        new_idxs[ax] = slice(None, None, -1)
    return tuple(new_idxs)


def grad_convolve(argnum, ans, A, B, axes=None, dot_axes=[(), ()], mode="full"):
    assert mode in ["valid", "full"], f"Grad for mode {mode} not yet implemented"
    axes, shapes = parse_axes(A.shape, B.shape, axes, dot_axes, mode)
    if argnum == 0:
        X, Y = A, B
        _X_, _Y_ = "A", "B"
        ignore_Y = "ignore_B"
    elif argnum == 1:
        X, Y = B, A
        _X_, _Y_ = "B", "A"
        ignore_Y = "ignore_A"
    else:
        raise NotImplementedError(f"Can't take grad of convolve w.r.t. arg {argnum}")

    if mode == "full":
        new_mode = "valid"
    else:
        if any([x_size > y_size for x_size, y_size in zip(shapes[_X_]["conv"], shapes[_Y_]["conv"])]):
            new_mode = "full"
        else:
            new_mode = "valid"

    def vjp(g):
        result = convolve(
            g,
            Y[flipped_idxs(Y.ndim, axes[_Y_]["conv"])],
            axes=[axes["out"]["conv"], axes[_Y_]["conv"]],
            dot_axes=[axes["out"][ignore_Y], axes[_Y_]["ignore"]],
            mode=new_mode,
        )
        new_order = npo.argsort(axes[_X_]["ignore"] + axes[_X_]["dot"] + axes[_X_]["conv"])
        return np.transpose(result, new_order)

    return vjp


defvjp(convolve, partial(grad_convolve, 0), partial(grad_convolve, 1))



import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad, make_vjp
from autograd.misc import const_graph, flatten
from autograd.test_util import scalar_close
from autograd.tracer import primitive


def test_const_graph():
    L = []

    def foo(x, y):
        L.append(None)
        return grad(lambda x: np.sin(x) + x * 2)(x * y)

    foo_wrapped = const_graph(foo)

    assert len(L) == 0
    assert scalar_close(foo(0.0, 0.0), foo_wrapped(0.0, 0.0))
    assert len(L) == 2
    assert scalar_close(foo(1.0, 0.5), foo_wrapped(1.0, 0.5))
    assert len(L) == 3
    assert scalar_close(foo(1.0, 0.5), foo_wrapped(1.0, 0.5))
    assert len(L) == 4


def test_const_graph_args():
    L = []

    @primitive
    def process(var, varname):
        L.append(varname)
        return var

    def foo(x, y, z):
        x = process(x, "x")
        y = process(y, "y")
        z = process(z, "z")
        return x + 2 * y + 3 * z

    foo_wrapped = const_graph(foo, 1.0, z=3.0)

    assert L == []
    assert scalar_close(foo(1.0, 2.0, 3.0), foo_wrapped(2.0))
    assert L == ["x", "y", "z", "x", "y", "z"]
    L = []
    assert scalar_close(foo(1.0, 2.0, 3.0), foo_wrapped(2.0))
    assert L == ["x", "y", "z", "y"]
    L = []
    assert scalar_close(foo(1.0, 2.0, 3.0), foo_wrapped(2.0))
    assert L == ["x", "y", "z", "y"]


def test_flatten():
    r = np.random.randn
    x = (1.0, r(2, 3), [r(1, 4), {"x": 2.0, "y": r(4, 2)}])
    x_flat, unflatten = flatten(x)
    assert x_flat.shape == (20,)
    assert x_flat[0] == 1.0
    assert np.all(x_flat == flatten(unflatten(x_flat))[0])

    y = (1.0, 2.0, [3.0, {"x": 2.0, "y": 4.0}])
    y_flat, unflatten = flatten(y)
    assert y_flat.shape == (5,)
    assert y == unflatten(y_flat)


def test_flatten_empty():
    val = (npr.randn(4), [npr.randn(3, 4), 2.5], (), (2.0, [1.0, npr.randn(2)]))
    vect, unflatten = flatten(val)
    val_recovered = unflatten(vect)
    vect_2, _ = flatten(val_recovered)
    assert np.all(vect == vect_2)


def test_flatten_dict():
    val = {"k": npr.random((4, 4)), "k2": npr.random((3, 3)), "k3": 3.0, "k4": [1.0, 4.0, 7.0, 9.0]}

    vect, unflatten = flatten(val)
    val_recovered = unflatten(vect)
    vect_2, _ = flatten(val_recovered)
    assert np.all(vect == vect_2)


def unflatten_tracing():
    val = [npr.randn(4), [npr.randn(3, 4), 2.5], (), (2.0, [1.0, npr.randn(2)])]
    vect, unflatten = flatten(val)

    def f(vect):
        return unflatten(vect)

    flatten2, _ = make_vjp(f)(vect)
    assert np.all(vect == flatten2(val))


def test_flatten_nodes_in_containers():
    # see issue #232
    def f(x, y):
        xy, _ = flatten([x, y])
        return np.sum(xy)

    grad(f)(1.0, 2.0)


def test_flatten_complex():
    val = 1 + 1j
    flat, unflatten = flatten(val)
    assert np.all(val == unflatten(flat))



"""Implements a Gaussian mixture model, in which parameters are fit using
gradient descent.  This example runs on 2-dimensional data, but the model
works on arbitrarily-high dimension."""

import matplotlib.pyplot as plt
from data import make_pinwheel
from scipy.optimize import minimize

import autograd.numpy as np
import autograd.numpy.random as npr
import autograd.scipy.stats.multivariate_normal as mvn
from autograd import grad, hessian_vector_product
from autograd.misc.flatten import flatten_func
from autograd.scipy.special import logsumexp


def init_gmm_params(num_components, D, scale, rs=npr.RandomState(0)):
    return {
        "log proportions": rs.randn(num_components) * scale,
        "means": rs.randn(num_components, D) * scale,
        "lower triangles": np.zeros((num_components, D, D)) + np.eye(D),
    }


def log_normalize(x):
    return x - logsumexp(x)


def unpack_gmm_params(params):
    normalized_log_proportions = log_normalize(params["log proportions"])
    return normalized_log_proportions, params["means"], params["lower triangles"]


def gmm_log_likelihood(params, data):
    cluster_lls = []
    for log_proportion, mean, cov_sqrt in zip(*unpack_gmm_params(params)):
        cov = np.dot(cov_sqrt.T, cov_sqrt)
        cluster_lls.append(log_proportion + mvn.logpdf(data, mean, cov))
    return np.sum(logsumexp(np.vstack(cluster_lls), axis=0))


def plot_ellipse(ax, mean, cov_sqrt, alpha, num_points=100):
    angles = np.linspace(0, 2 * np.pi, num_points)
    circle_pts = np.vstack([np.cos(angles), np.sin(angles)]).T * 2.0
    cur_pts = mean + np.dot(circle_pts, cov_sqrt)
    ax.plot(cur_pts[:, 0], cur_pts[:, 1], "-", alpha=alpha)


def plot_gaussian_mixture(params, ax):
    for log_proportion, mean, cov_sqrt in zip(*unpack_gmm_params(params)):
        alpha = np.minimum(1.0, np.exp(log_proportion) * 10)
        plot_ellipse(ax, mean, cov_sqrt, alpha)


if __name__ == "__main__":
    init_params = init_gmm_params(num_components=10, D=2, scale=0.1)

    data = make_pinwheel(radial_std=0.3, tangential_std=0.05, num_classes=3, num_per_class=100, rate=0.4)

    def objective(params):
        return -gmm_log_likelihood(params, data)

    flattened_obj, unflatten, flattened_init_params = flatten_func(objective, init_params)

    fig = plt.figure(figsize=(12, 8), facecolor="white")
    ax = fig.add_subplot(111, frameon=False)
    plt.show(block=False)

    def callback(flattened_params):
        params = unflatten(flattened_params)
        print(f"Log likelihood {-objective(params)}")
        ax.cla()
        ax.plot(data[:, 0], data[:, 1], "k.")
        ax.set_xticks([])
        ax.set_yticks([])
        plot_gaussian_mixture(params, ax)
        plt.draw()
        plt.pause(1.0 / 60.0)

    minimize(
        flattened_obj,
        flattened_init_params,
        jac=grad(flattened_obj),
        hessp=hessian_vector_product(flattened_obj),
        method="Newton-CG",
        callback=callback,
    )



from functools import partial

import numpy as npo

try:
    import scipy
except:
    from warnings import warn

    warn("Skipping scipy tests.")
else:
    from numpy_utils import unary_ufunc_check
    from scipy.signal import convolve as sp_convolve

    import autograd.numpy as np
    import autograd.numpy.random as npr
    import autograd.scipy.integrate as integrate
    import autograd.scipy.linalg as spla
    import autograd.scipy.signal
    import autograd.scipy.special as special
    import autograd.scipy.stats as stats
    import autograd.scipy.stats.multivariate_normal as mvn
    from autograd import grad
    from autograd.test_util import check_grads, combo_check

    npr.seed(1)
    R = npr.randn
    U = npr.uniform

    # Fwd mode not yet implemented for scipy functions
    combo_check = partial(combo_check, modes=["rev"])
    unary_ufunc_check = partial(unary_ufunc_check, modes=["rev"])
    check_grads = partial(check_grads, modes=["rev"])

    def symmetrize_matrix_arg(fun, argnum):
        def T(X):
            return np.swapaxes(X, -1, -2) if np.ndim(X) > 1 else X

        def symmetrize(X):
            return 0.5 * (X + T(X))

        def symmetrized_fun(*args, **kwargs):
            args = list(args)
            args[argnum] = symmetrize(args[argnum])
            return fun(*args, **kwargs)

        return symmetrized_fun

    ### Stats ###
    def test_chi2_pdf():
        combo_check(stats.chi2.pdf, [0])([R(4) ** 2 + 1.1], [1, 2, 3])

    def test_chi2_cdf():
        combo_check(stats.chi2.cdf, [0])([R(4) ** 2 + 1.1], [1, 2, 3])

    def test_chi2_logpdf():
        combo_check(stats.chi2.logpdf, [0])([R(4) ** 2 + 1.1], [1, 2, 3])

    def test_beta_cdf():
        combo_check(stats.beta.cdf, [0])([U(0.0, 1.0, 4)], [R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1])

    def test_beta_pdf():
        combo_check(stats.beta.pdf, [0, 1, 2])([U(0.0, 1.0, 4)], [R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1])

    def test_beta_logpdf():
        combo_check(stats.beta.logpdf, [0, 1, 2])([U(0.0, 1.0, 4)], [R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1])

    def test_gamma_cdf():
        combo_check(stats.gamma.cdf, [0])([R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1])

    def test_gamma_pdf():
        combo_check(stats.gamma.pdf, [0, 1])([R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1])

    def test_gamma_logpdf():
        combo_check(stats.gamma.logpdf, [0, 1])([R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1])

    def test_norm_pdf():
        combo_check(stats.norm.pdf, [0, 1, 2])([R(4)], [R(4)], [R(4) ** 2 + 1.1])

    def test_norm_cdf():
        combo_check(stats.norm.cdf, [0, 1, 2])([R(4)], [R(4)], [R(4) ** 2 + 1.1])

    def test_norm_sf():
        combo_check(stats.norm.sf, [0, 1, 2])([R(4)], [R(4)], [R(4) ** 2 + 1.1])

    def test_norm_logpdf():
        combo_check(stats.norm.logpdf, [0, 1, 2])([R(4)], [R(4)], [R(4) ** 2 + 1.1])

    def test_norm_logcdf():
        combo_check(stats.norm.logcdf, [0, 1, 2])([R(4)], [R(4)], [R(4) ** 2 + 1.1])

    def test_norm_logsf():
        combo_check(stats.norm.logsf, [0, 1, 2])([R(4)], [R(4)], [R(4) ** 2 + 1.1])

    def test_norm_pdf_broadcast():
        combo_check(stats.norm.pdf, [0, 1, 2])([R(4, 3)], [R(1, 3)], [R(4, 1) ** 2 + 1.1])

    def test_norm_cdf_broadcast():
        combo_check(stats.norm.cdf, [0, 1, 2])([R(4, 3)], [R(1, 3)], [R(4, 1) ** 2 + 1.1])

    def test_norm_sf_broadcast():
        combo_check(stats.norm.cdf, [0, 1, 2])([R(4, 3)], [R(1, 3)], [R(4, 1) ** 2 + 1.1])

    def test_norm_logpdf_broadcast():
        combo_check(stats.norm.logpdf, [0, 1, 2])([R(4, 3)], [R(1, 3)], [R(4, 1) ** 2 + 1.1])

    def test_norm_logcdf_broadcast():
        combo_check(stats.norm.logcdf, [0, 1, 2])([R(4, 3)], [R(1, 3)], [R(4, 1) ** 2 + 1.1])

    def test_norm_logsf_broadcast():
        combo_check(stats.norm.logcdf, [0, 1, 2])([R(4, 3)], [R(1, 3)], [R(4, 1) ** 2 + 1.1])

    def test_poisson_cdf():
        combo_check(stats.poisson.cdf, [1])([np.round(R(4) ** 2)], [R(4) ** 2 + 1.1])

    def test_poisson_logpmf():
        combo_check(stats.poisson.logpmf, [1])([np.round(R(4) ** 2)], [R(4) ** 2 + 1.1])

    def test_poisson_pmf():
        combo_check(stats.poisson.pmf, [1])([np.round(R(4) ** 2)], [R(4) ** 2 + 1.1])

    def test_poisson_cdf_broadcast():
        combo_check(stats.poisson.cdf, [1])([np.round(R(4, 3) ** 2)], [R(4, 1) ** 2 + 1.1])

    def test_poisson_logpmf_broadcast():
        combo_check(stats.poisson.logpmf, [1])([np.round(R(4, 3) ** 2)], [R(4, 1) ** 2 + 1.1])

    def test_poisson_pmf_broadcast():
        combo_check(stats.poisson.pmf, [1])([np.round(R(4, 3) ** 2)], [R(4, 1) ** 2 + 1.1])

    def test_t_pdf():
        combo_check(stats.t.pdf, [0, 1, 2, 3])([R(4)], [R(4) ** 2 + 2.1], [R(4)], [R(4) ** 2 + 2.1])

    def test_t_cdf():
        combo_check(stats.t.cdf, [0, 2])([R(4)], [R(4) ** 2 + 2.1], [R(4)], [R(4) ** 2 + 2.1])

    def test_t_logpdf():
        combo_check(stats.t.logpdf, [0, 1, 2, 3])([R(4)], [R(4) ** 2 + 2.1], [R(4)], [R(4) ** 2 + 2.1])

    def test_t_logcdf():
        combo_check(stats.t.logcdf, [0, 2])([R(4)], [R(4) ** 2 + 2.1], [R(4)], [R(4) ** 2 + 2.1])

    def test_t_pdf_broadcast():
        combo_check(stats.t.pdf, [0, 1, 2, 3])(
            [R(4, 3)], [R(1, 3) ** 2 + 2.1], [R(4, 3)], [R(4, 1) ** 2 + 2.1]
        )

    def test_t_cdf_broadcast():
        combo_check(stats.t.cdf, [0, 2])([R(4, 3)], [R(1, 3) ** 2 + 2.1], [R(4, 3)], [R(4, 1) ** 2 + 2.1])

    def test_t_logpdf_broadcast():
        combo_check(stats.t.logpdf, [0, 1, 2, 3])(
            [R(4, 3)], [R(1, 3) ** 2 + 2.1], [R(4, 3)], [R(4, 1) ** 2 + 2.1]
        )

    def test_t_logcdf_broadcast():
        combo_check(stats.t.logcdf, [0, 2])([R(4, 3)], [R(1, 3) ** 2 + 2.1], [R(4, 3)], [R(4, 1) ** 2 + 2.1])

    def make_psd(mat):
        return np.dot(mat.T, mat) + np.eye(mat.shape[0])

    def test_mvn_pdf():
        combo_check(symmetrize_matrix_arg(mvn.pdf, 2), [0, 1, 2])(
            [R(4)], [R(4)], [make_psd(R(4, 4))], allow_singular=[False]
        )

    def test_mvn_logpdf():
        combo_check(symmetrize_matrix_arg(mvn.logpdf, 2), [0, 1, 2])(
            [R(4)], [R(4)], [make_psd(R(4, 4))], allow_singular=[False]
        )

    def test_mvn_entropy():
        combo_check(symmetrize_matrix_arg(mvn.entropy, 1), [0, 1])([10 * R(4)], [make_psd(R(4, 4))])

    def test_mvn_sing_cov():
        cov = np.zeros((4, 4))
        cov[0, 0] = cov[1, 1] = 1

        # Only allow variations in x along the first two dimensions, because
        # variance is zero in the last two.
        def pdf(x, mean, cov):
            x = np.concatenate([x[:2], mean[2:]])
            return symmetrize_matrix_arg(partial(mvn.pdf, allow_singular=True), 2)(x, mean, cov)

        combo_check(pdf, [0, 1])(
            [np.concatenate((R(2), np.zeros(2)))], [np.concatenate((R(2), np.zeros(2)))], [cov]
        )

        def logpdf(x, mean, cov):
            x = np.concatenate([x[:2], mean[2:]])
            return symmetrize_matrix_arg(partial(mvn.logpdf, allow_singular=True), 2)(x, mean, cov)

        combo_check(logpdf, [0, 1])(
            [np.concatenate((R(2), np.zeros(2)))], [np.concatenate((R(2), np.zeros(2)))], [cov]
        )

    def test_mvn_pdf_broadcast():
        combo_check(symmetrize_matrix_arg(mvn.pdf, 2), [0, 1, 2])([R(5, 4)], [R(4)], [make_psd(R(4, 4))])

    def test_mvn_logpdf_broadcast():
        combo_check(symmetrize_matrix_arg(mvn.logpdf, 2), [0, 1, 2])([R(5, 4)], [R(4)], [make_psd(R(4, 4))])

    alpha = npr.random(4) ** 2 + 1.2
    x = stats.dirichlet.rvs(alpha, size=1)[0, :]

    # Need to normalize input so that x's sum to one even when we perturb them to compute numeric gradient.
    def normalize(x):
        return x / sum(x)

    def normalized_dirichlet_pdf(x, alpha):
        return stats.dirichlet.pdf(normalize(x), alpha)

    def normalized_dirichlet_logpdf(x, alpha):
        return stats.dirichlet.logpdf(normalize(x), alpha)

    def test_dirichlet_pdf_x():
        combo_check(normalized_dirichlet_pdf, [0])([x], [alpha])

    def test_dirichlet_pdf_alpha():
        combo_check(stats.dirichlet.pdf, [1])([x], [alpha])

    def test_dirichlet_logpdf_x():
        combo_check(normalized_dirichlet_logpdf, [0])([x], [alpha])

    def test_dirichlet_logpdf_alpha():
        combo_check(stats.dirichlet.logpdf, [1])([x], [alpha])

    ### Misc ###
    def test_logsumexp1():
        combo_check(special.logsumexp, [0], modes=["fwd", "rev"])(
            [np.array([1.1]), R(4), R(3, 4)], axis=[None, 0], keepdims=[True, False]
        )

    def test_logsumexp2():
        combo_check(special.logsumexp, [0], modes=["fwd", "rev"])(
            [R(3, 4), R(4, 5, 6), R(1, 5)], axis=[None, 0, 1], keepdims=[True, False]
        )

    def test_logsumexp3():
        combo_check(special.logsumexp, [0], modes=["fwd", "rev"])(
            [R(4)], b=[np.exp(R(4))], axis=[None, 0], keepdims=[True, False]
        )

    def test_logsumexp4():
        combo_check(special.logsumexp, [0], modes=["fwd", "rev"])(
            [
                R(3, 4),
            ],
            b=[np.exp(R(3, 4))],
            axis=[None, 0, 1],
            keepdims=[True, False],
        )

    def test_logsumexp5():
        combo_check(special.logsumexp, [0], modes=["fwd", "rev"])(
            [R(2, 3, 4)], b=[np.exp(R(2, 3, 4))], axis=[None, 0, 1], keepdims=[True, False]
        )

    def test_logsumexp6():
        x = npr.randn(1, 5)

        def f(a):
            return special.logsumexp(a, axis=1, keepdims=True)

        check_grads(f, modes=["fwd", "rev"])(x)
        check_grads(lambda a: grad(f)(a), modes=["fwd", "rev"])(x)

    ### Signal ###
    def test_convolve_generalization():
        ag_convolve = autograd.scipy.signal.convolve
        A_35 = R(3, 5)
        A_34 = R(3, 4)
        A_342 = R(3, 4, 2)
        A_2543 = R(2, 5, 4, 3)
        A_24232 = R(2, 4, 2, 3, 2)

        for mode in ["valid", "full"]:
            assert npo.allclose(
                ag_convolve(A_35, A_34, axes=([1], [0]), mode=mode)[1, 2],
                sp_convolve(A_35[1, :], A_34[:, 2], mode),
            )
            assert npo.allclose(
                ag_convolve(A_35, A_34, axes=([], []), dot_axes=([0], [0]), mode=mode),
                npo.tensordot(A_35, A_34, axes=([0], [0])),
            )
            assert npo.allclose(
                ag_convolve(A_35, A_342, axes=([1], [2]), dot_axes=([0], [0]), mode=mode)[2],
                sum([sp_convolve(A_35[i, :], A_342[i, 2, :], mode) for i in range(3)]),
            )
            assert npo.allclose(
                ag_convolve(A_2543, A_24232, axes=([1, 2], [2, 4]), dot_axes=([0, 3], [0, 3]), mode=mode)[2],
                sum(
                    [
                        sum(
                            [sp_convolve(A_2543[i, :, :, j], A_24232[i, 2, :, j, :], mode) for i in range(2)]
                        )
                        for j in range(3)
                    ]
                ),
            )

    def test_convolve():
        combo_check(autograd.scipy.signal.convolve, [0, 1])(
            [R(4), R(5), R(6)], [R(2), R(3), R(4)], mode=["full", "valid"]
        )

    def test_convolve_2d():
        combo_check(autograd.scipy.signal.convolve, [0, 1])(
            [R(4, 3), R(5, 4), R(6, 7)], [R(2, 2), R(3, 2), R(4, 2), R(4, 1)], mode=["full", "valid"]
        )

    def test_convolve_ignore():
        combo_check(autograd.scipy.signal.convolve, [0, 1])(
            [R(4, 3)],
            [R(3, 2)],
            axes=[([0], [0]), ([1], [1]), ([0], [1]), ([1], [0]), ([0, 1], [0, 1]), ([1, 0], [1, 0])],
            mode=["full", "valid"],
        )

    def test_convolve_ignore_dot():
        combo_check(autograd.scipy.signal.convolve, [0, 1])(
            [R(3, 3, 2)],
            [R(3, 2, 3)],
            axes=[([1], [1])],
            dot_axes=[([0], [2]), ([0], [0])],
            mode=["full", "valid"],
        )

    ### Special ###
    def test_beta():
        combo_check(special.beta, [0, 1])([R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1])

    def test_betainc():
        combo_check(special.betainc, [2])([R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1], [U(0.0, 1.0, 4)])

    def test_betaln():
        combo_check(special.betaln, [0, 1])([R(4) ** 2 + 1.1], [R(4) ** 2 + 1.1])

    def test_gammainc():
        combo_check(special.gammainc, [1])([1], R(4) ** 2 + 1.3)

    def test_gammaincc():
        combo_check(special.gammaincc, [1])([1], R(4) ** 2 + 1.3)

    def test_polygamma():
        combo_check(special.polygamma, [1])([0], R(4) ** 2 + 1.3)

    def test_jn():
        combo_check(special.jn, [1])([2], R(4) ** 2 + 1.3)

    def test_yn():
        combo_check(special.yn, [1])([2], R(4) ** 2 + 1.3)

    def test_psi():
        unary_ufunc_check(special.psi, lims=[0.3, 2.0], test_complex=False)

    def test_digamma():
        unary_ufunc_check(special.digamma, lims=[0.3, 2.0], test_complex=False)

    def test_gamma():
        unary_ufunc_check(special.gamma, lims=[0.3, 2.0], test_complex=False)

    def test_gammaln():
        unary_ufunc_check(special.gammaln, lims=[0.3, 2.0], test_complex=False)

    def test_gammasgn():
        unary_ufunc_check(special.gammasgn, lims=[0.3, 2.0], test_complex=False)

    def test_rgamma():
        unary_ufunc_check(special.rgamma, lims=[0.3, 2.0], test_complex=False)

    def test_multigammaln():
        combo_check(special.multigammaln, [0])([U(4.0, 5.0), U(4.0, 5.0, (2, 3))], [1, 2, 3])

    def test_j0():
        unary_ufunc_check(special.j0, lims=[0.2, 20.0], test_complex=False)

    def test_j1():
        unary_ufunc_check(special.j1, lims=[0.2, 20.0], test_complex=False)

    def test_y0():
        unary_ufunc_check(special.y0, lims=[0.2, 20.0], test_complex=False)

    def test_y1():
        unary_ufunc_check(special.y1, lims=[0.2, 20.0], test_complex=False)

    def test_i0():
        unary_ufunc_check(special.i0, lims=[0.2, 20.0], test_complex=False)

    def test_i1():
        unary_ufunc_check(special.i1, lims=[0.2, 20.0], test_complex=False)

    def test_iv():
        combo_check(special.iv, [1])(U(1.0, 50.0, 4), R(4) ** 2 + 1.3)

    def test_ive():
        combo_check(special.ive, [1])(U(1.0, 50.0, 4), R(4) ** 2 + 1.3)

    def test_erf():
        unary_ufunc_check(special.erf, lims=[-3.0, 3.0], test_complex=True)

    def test_erfc():
        unary_ufunc_check(special.erfc, lims=[-3.0, 3.0], test_complex=True)

    def test_erfinv():
        unary_ufunc_check(special.erfinv, lims=[-0.95, 0.95], test_complex=False)

    def test_erfcinv():
        unary_ufunc_check(special.erfcinv, lims=[0.05, 1.95], test_complex=False)

    def test_logit():
        unary_ufunc_check(special.logit, lims=[0.10, 0.90], test_complex=False)

    def test_expit():
        unary_ufunc_check(special.expit, lims=[-4.05, 4.95], test_complex=False)

    ### ODE integrator ###
    def func(y, t, arg1, arg2):
        return -np.sqrt(t) - y + arg1 - np.mean((y + arg2) ** 2)

    def test_odeint():
        combo_check(integrate.odeint, [1, 2, 3])([func], [R(3)], [np.linspace(0.1, 0.2, 4)], [(R(3), R(3))])

    ## Linalg
    def test_sqrtm():
        combo_check(spla.sqrtm, modes=["fwd"], order=2)([R(3, 3)])

    def test_sqrtm():
        combo_check(symmetrize_matrix_arg(spla.sqrtm, 0), modes=["fwd", "rev"], order=2)([R(3, 3)])

    def test_solve_sylvester():
        combo_check(spla.solve_sylvester, [0, 1, 2], modes=["rev", "fwd"], order=2)(
            [R(3, 3)], [R(3, 3)], [R(3, 3)]
        )

    def test_solve_banded():
        combo_check(spla.solve_banded, [1, 2], modes=["rev"], order=1)([(1, 1)], [R(3, 5)], [R(5)])



import matplotlib.pyplot as plt

# same BBSVI function!
from black_box_svi import black_box_variational_inference

import autograd.numpy as np
import autograd.scipy.stats.norm as norm
from autograd.misc.optimizers import adam, sgd

if __name__ == "__main__":
    # Specify an inference problem by its unnormalized log-density.
    # it's difficult to see the benefit in low dimensions
    # model parameters are a mean and a log_sigma
    np.random.seed(42)
    obs_dim = 20
    Y = np.random.randn(obs_dim, obs_dim).dot(np.random.randn(obs_dim))

    def log_density(x, t):
        mu, log_sigma = x[:, :obs_dim], x[:, obs_dim:]
        sigma_density = np.sum(norm.logpdf(log_sigma, 0, 1.35), axis=1)
        mu_density = np.sum(norm.logpdf(Y, mu, np.exp(log_sigma)), axis=1)
        return sigma_density + mu_density

    # Build variational objective.
    D = obs_dim * 2  # dimension of our posterior
    objective, gradient, unpack_params = black_box_variational_inference(log_density, D, num_samples=2000)

    # Define the natural gradient
    #   The natural gradient of the ELBO is the gradient of the elbo,
    #   preconditioned by the inverse Fisher Information Matrix.  The Fisher,
    #   in the case of a diagonal gaussian, is a diagonal matrix that is a
    #   simple function of the variance.  Intuitively, statistical distance
    #   created by perturbing the mean of an independent Gaussian is
    #   determined by how wide the distribution is along that dimension ---
    #   the wider the distribution, the less sensitive statistical distances is
    #   to perturbations of the mean; the narrower the distribution, the more
    #   the statistical distance changes when you perturb the mean (imagine
    #   an extremely narrow Gaussian --- basically a spike.  The KL between
    #   this Gaussian and a Gaussian $\epsilon$ away in location can be big ---
    #   moving the Gaussian could significantly reduce overlap in support
    #   which corresponds to a greater statistical distance).
    #
    #   When we want to move in directions of steepest ascent, we multiply by
    #   the inverse fisher --- that way we make quicker progress when the
    #   variance is wide, and we scale down our step size when the variance
    #   is small (which leads to more robust/less chaotic ascent).
    def fisher_diag(lam):
        mu, log_sigma = unpack_params(lam)
        return np.concatenate([np.exp(-2.0 * log_sigma), np.ones(len(log_sigma)) * 2])

    # simple! basically free!
    natural_gradient = lambda lam, i: (1.0 / fisher_diag(lam)) * gradient(lam, i)

    # function for keeping track of callback ELBO values (for plotting below)
    def optimize_and_lls(optfun):
        num_iters = 200
        elbos = []

        def callback(params, t, g):
            elbo_val = -objective(params, t)
            elbos.append(elbo_val)
            if t % 50 == 0:
                print(f"Iteration {t} lower bound {elbo_val}")

        init_mean = -1 * np.ones(D)
        init_log_std = -5 * np.ones(D)
        init_var_params = np.concatenate([init_mean, init_log_std])
        variational_params = optfun(num_iters, init_var_params, callback)
        return np.array(elbos)

    # let's optimize this with a few different step sizes
    elbo_lists = []
    step_sizes = [0.1, 0.25, 0.5]
    for step_size in step_sizes:
        # optimize with standard gradient + adam
        optfun = lambda n, init, cb: adam(gradient, init, step_size=step_size, num_iters=n, callback=cb)
        standard_lls = optimize_and_lls(optfun)

        # optimize with natural gradient + sgd, no momentum
        optnat = lambda n, init, cb: sgd(
            natural_gradient, init, step_size=step_size, num_iters=n, callback=cb, mass=0.001
        )
        natural_lls = optimize_and_lls(optnat)
        elbo_lists.append((standard_lls, natural_lls))

    # visually compare the ELBO
    plt.figure(figsize=(12, 8))
    colors = ["b", "k", "g"]
    for col, ss, (stand_lls, nat_lls) in zip(colors, step_sizes, elbo_lists):
        plt.plot(
            np.arange(len(stand_lls)),
            stand_lls,
            "--",
            label="standard (adam, step-size = %2.2f)" % ss,
            alpha=0.5,
            c=col,
        )
        plt.plot(np.arange(len(nat_lls)), nat_lls, "-", label="natural (sgd, step-size = %2.2f)" % ss, c=col)

    llrange = natural_lls.max() - natural_lls.min()
    plt.ylim((natural_lls.max() - llrange * 0.1, natural_lls.max() + 10))
    plt.xlabel("optimization iteration")
    plt.ylabel("ELBO")
    plt.legend(loc="lower right")
    plt.title("%d dimensional posterior" % D)
    plt.show()
